/*
 * tc.c - anonymous direct chat over Tor in PGP - fausto@ngi.it
 * <r4f5fajp2oiuhjqyass3nogu55sivxnc42fnwankiynf6o5jpno6exad.onion/tc.html>
 *
 *      I'm not a person who particularly had heros when growing up
 *
 *                                            -- Dennis Ritchie
 */

#include <syslog.h>
#include <stdarg.h>
#include <err.h>
#include <errno.h>
#include <assert.h>
#include <stdlib.h>
#ifdef openbsd
#include <string.h>
#endif
#ifdef linux
#include <bsd/string.h>
#endif
#include <unistd.h>
#include <stdio.h>
#include <ctype.h>
#include <curses.h>
#include <time.h>
#include <getopt.h>
#include <signal.h>
#include <pthread.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#define TCURL "r4f5fajp2oiuhjqyass3nogu55sivxnc42fnwankiynf6o5jpno6exad.onion"
#define MAX_CONTACTS 100           /* contacts limit                  */
#define MAX_GROUPS   50            /* groups limit                    */
#define MAX_MEMBERS  5             /* members per group limit         */
#define MAX_SNDFSIZE 3000          /* file size byte limit            */
#define MAX_FSIZE    11000         /* tc.conf file size limit         */
#define N            10            /* nickname lenght and some limits */
#define NFILE        19            /* files loaded by localfilez()    */
#define COMMANDS     39            /* commands                        */
#define FLAGS        12            /* flags                           */
#define ALIVE        30            /* Tor keep alive interval         */
#define IN           60            /* input text lenght               */
#define BACKLOG      64            /* pending connections queue hold  */
#define MSG          75            /* multirole string message & more */
#define STR          200           /* multirole string system & more  */
#define RWLENGTH     256           /* length fread/fwrite             */
#define SKA          8             /* symmetric-key algorithms        */

struct {

    char name[N];                  /* contact's name in tc.conf       */
    char key[41];                  /* contact's pgp key id            */
    char onion[63];                /* contact's onion addess          */

} c[MAX_CONTACTS];

struct {

    char grpname[N];               /* group name                      */
    char members[MAX_MEMBERS][N];  /* group members                   */

} g[MAX_GROUPS];

struct {

    char keyid[41];                /* pgp key ids                     */
    char descr[70];                /* pgp key description mail        */

} ring[MAX_CONTACTS];

struct {

    char onion[64];                /* saved onion address             */

} restore[N];

struct {

    char auth[STR];                /* .auth files                     */

} aauth[N];

struct {

    char prv[STR];                 /* .auth_private files             */

} aprv[N];

struct tm tm;                      /* time                            */
time_t tstart, tstop, t;           /* tc uptime and time              */
pthread_t clk, srv, dec;           /* clocker listener send_decoy     */
pthread_mutex_t mtxtct;            /* lock TCT write in ucrypt/gcrypt */
WINDOW *mainwin, *txt, *pad;       /* mainwin, text and pad scrolling */
pid_t tpid = 0;                    /* Tor process id                  */
long bytein, byteout;              /* tc cyphered traffic counters    */
const char *home, *jail, *trm;     /* const because unveil()          */

int TPORT, TCPORT, burst, nmex, ndec;
short decoy, lazy, logs, voice, sym, away, tor, flood, con, grp, dre, x, y,
      noblank, scrollback, nusr, ngrp, nkey, nrow, who, hpos, row, col, info_f,
      info_b, bars_f, bars_b, text_f, text_b, erro_f, erro_b, interval, strict,
      nauth, nprv, camuflage, symg, symo;
char *TCS, *TCT, *TCR, *TCN, *tcconf, *tclog, *tclock, *tcbot, *tcbotsh, *tcdir,
     *tcrcvfile, *tcrc, *tchost, *tcauthdir, *tchiddendir, *tcauthclidir,
     *tctordir, *tcver, *tctgz, *k[5], *o[5], chk[17], chktor[17], K[MSG],
     in[MSG], msg[MSG], way[MSG], zzz[21], nick[N], remo[N], flag[12],
     PASS[MSG], TPROXY[16], TCDAEMON[16], tport[6], s[MSG][MSG], socks4a[STR],
     callo[63], callk[63], gpgver[N], torver[N], cipher[12];
char tag[]  = "tc - anonymous direct chat over Tor in PGP";
char *commands[COMMANDS] = {"/algo","/away","/bot","/call","/camuflage",
     "/circuit","/clear","/colors","/date","/decoy","/disconnect","/exit",
     "/flags","/flood","/generate","/groups","/help","/history","/info","/join",
     "/keys","/lazy","/log","/me","/names","/nick","/part","/quit","/restore",
     "/send","/show","/strict","/symmetric","/topic","/uptime","/voice","/who",
     "/whois",0};
char *flags[FLAGS] = {"[D] decoy","[L] log","[V] voice","[S] symmetric",
     "[A] away", "[T] tor","[F] antiflood","[Z] lazyness","[X] authorize",
     "[C] camuflage","[G] group",0};
char *alg[SKA] = {"AES","AES192","AES256","CAMELLIA128","CAMELLIA192",
     "CAMELLIA256","TWOFISH",0};
char *colors[9] = {"BLACK","RED","GREEN","YELLOW","BLUE","MAGENTA","CYAN",
     "WHITE",0};
void c_info(void) {init_pair(1,info_f,info_b);
     wattron(txt, A_BOLD | COLOR_PAIR(1));}
void c_info_off(void) {wattroff(txt, A_BOLD | COLOR_PAIR(1));}
void c_bars(void) {init_pair(2,bars_f,bars_b); color_set(2,NULL);}
void c_text(void) {init_pair(3,text_f,text_b);
     wattron(txt, A_NORMAL | COLOR_PAIR(3));}
void c_text_off(void) {wattroff(txt, A_NORMAL | COLOR_PAIR(3));}
void c_erro(void) {init_pair(4,erro_f,erro_b);
     wattron(txt, A_BOLD | COLOR_PAIR(4));}
void c_erro_off(void) {wattroff(txt, A_BOLD | COLOR_PAIR(4));}
void c_whit(void) {init_pair(5,7,0); wattron(txt, A_BOLD | COLOR_PAIR(5));}
void c_whit_off(void) {wattroff(txt, A_BOLD | COLOR_PAIR(5));}
void c_prompt(void) {init_pair(6,text_f,text_b); color_set(6,NULL);}
void c_oldtrm(void) {init_pair(7,0,7);
     wattron(mainwin, WA_REVERSE | COLOR_PAIR(7));}
void c_oldtrm_off(void) {wattroff(mainwin, WA_REVERSE | COLOR_PAIR(7));}
void hm(void) { wprintw(txt, "%02d:%02d ", tm.tm_hour, tm.tm_min); }

void colorinfo(void) {

    hm(); c_info(); wprintw(txt, "<*> "); c_info_off();
}

void colorerror(void) {

    hm(); c_erro(); wprintw(txt, "<!> "); c_erro_off();
}

void usage(void) {

    errx(1, "[-n nickname] [-s nickname] [-u]");
}

void waitingfor(void) {

    curs_set(0); move(row-1, 0); clrtoeol();
    mvprintw((row-1),0, " =--> [*] wait...");
    wrefresh(mainwin);
}

void help(void) {

    hm(); c_info(); wprintw(txt, "<*> commands:"); c_info_off();
    for(int i = 0; i < (COMMANDS-1); i++) {

        if(i % 5 == 0) wprintw(txt, "\n%02d:%02d ", tm.tm_hour, tm.tm_min);
        wprintw(txt, "%-14s", commands[i]);
    }
    wprintw(txt, "\n"); hm(); wprintw(txt, "(total count: %d)\n", COMMANDS-1);
}

void flagz(void) {

    short tic[FLAGS];
    tic[0] = decoy;  tic[1] = logs;      tic[2] = voice; tic[3] = sym;
    tic[4] = away;   tic[5] = tor;       tic[6] = flood; tic[7] = lazy;
    tic[8] = strict; tic[9] = camuflage; tic[10]= grp;

    hm(); c_info(); wprintw(txt, "<*> flags:"); c_info_off();
    for(int i = 0; i < (FLAGS-1); i++) {

        if(i % 5 == 0) wprintw(txt, "\n%02d:%02d ", tm.tm_hour, tm.tm_min);
        if(tic[i] == 1) c_whit();
        wprintw(txt, "%-14s", flags[i]); c_whit_off();
    }
    wprintw(txt, "\n"); hm(); wprintw(txt, "(total count: %d)\n", FLAGS-1);
}

int logger(char *nn, char *mm) {

    if(strstr(mm, " activated symmetric ") != NULL) return(1);
    if(nn == NULL) {

        FILE *f = fopen(tclog, "a");
        fprintf(f, "%02d:%02d %s\n", tm.tm_hour, tm.tm_min, mm);
        fclose(f);
    }
    else {

        FILE *f = fopen(tclog, "a");
        fprintf(f, "%02d:%02d <%s> %s\n",tm.tm_hour,tm.tm_min,nn,mm);
        fclose(f);
    }
    return(0);
}

char const *mkfake(int camuf) {
    /* ripped from ~el8[1] zine ahahah */
    if(camuf == 0) {

    char *name = (char *) calloc(1, 10);
    if(name == NULL) { free(name); return NULL; }
    *(name + 0) = 'A' + (lrand48() % ('Z' - 'A'));
    *(name + 1) = 'a' + (lrand48() % ('z' - 'a'));
    *(name + 2) = 'A' + (lrand48() % ('Z' - 'A'));
    *(name + 3) = 'a' + (lrand48() % ('z' - 'a'));
    *(name + 4) = 'A' + (lrand48() % ('Z' - 'A'));
    *(name + 5) = 'a' + (lrand48() % ('z' - 'a'));
    *(name + 6) = 'A' + (lrand48() % ('Z' - 'A'));
    *(name + 7) = 'a' + (lrand48() % ('z' - 'a'));
    *(name + 8) = 'A' + (lrand48() % ('Z' - 'A'));
    *(name + 9) = 'a' + (lrand48() % ('z' - 'a'));
    *(name + 10) = 0;
    return name;
    }
    if(camuf == 1) {

        int n = 19;
        char const *str[] = { /* push different crap anytime */
                        "....'.WB..E...C..@.@.x.......0.pool.ntp.org",
                        "..........UD....abcdefghijklmnopqrstuvwabcdefghi",
                        "%PDF-1.5.%......779 0 obj.<</Linearized 1/L 5187",
                        "......JFIF.............C........................",
                        "... ftypisom....isomiso2avc1mp41....freeJ...mdat",
                        "GIF89a..>...1...$..H..l...............$.$$.H$.l$",
                        "@charset utf-8;  /* Reset CSS */  body, h1, h2, ",
                        "/*! jQuery v@1.8.0 jquery.com | jquery.org/licen",
                        ".",
                        "..",
                        ".<DOCTYPE html>.",
                        "...F...",
                        "...x..i#ZZ9?.",
                        "font-size: 30px;",
                        "text-align:center; border:none; border-radius:5px;",
                        "*//*! normalize.css v3.0.3 | MIT License | ",
                        ",up.each({scrollLeft:pageXOffset,scrollTop:page11",
                        ".zcss(a,boxSizing)===border-box;..",
                        0
                       };
        int vs = lrand48() % (n-1);
        return str[vs];
    }
    return NULL;
}

void shred(char *file) {

    FILE *f = fopen(file, "w");
    if(f != NULL) {

        char buf[MAX_SNDFSIZE] = {"\0"};
        for(int j = 0; j < 300; j++) strlcat(buf, mkfake(0), sizeof(buf));
        fprintf(f, "%s", buf);
        fflush(f); fclose(f);
    }
    unlink(file);
}

void nocom(void) {

    if(decoy == 1) {

        pthread_cancel(dec); decoy = 0;
    }
    unlink(TCS); shred(TCT); unlink(TCR); shred(TCN);
    con = 0; grp = 0; who = 0; symg = 0; symo = 0;
    memset(remo, 0, sizeof(remo)); memset(k, 0, sizeof(k));
    memset(o, 0, sizeof(o)); memset(zzz, 0, sizeof(zzz));
    colorerror(); wprintw(txt, "not connected\n");
}

void topic(char *m) {

    strlcpy(tag, m, (N*3)+1);
}

void hexdump(unsigned char const *data) {

    for(int i = 0; i < 100; i += 16) {

        hm(); wprintw(txt, "%04x: ", i);
        for (int j = 0; j < 16; j++) {

            wprintw(txt,"%02x ", data[i+j] & 0xff);
        }
        wprintw(txt, "|");
        for (int z = 0; z < 16; z++) {

            int hex = data[i+z] & 0xff;
            if(!isprint(hex)) hex = '.'; wprintw(txt, "%c", hex);
        }
        wprintw(txt, "|\n");
    }
}

void draw(void) {

    char bar[400] = "";
    for(int b = 0; b < col; b++) bar[b] = ' ';
    if(strcmp("vt100", trm) == 0 || strcmp("vt220", trm ) == 0) c_oldtrm();
    else c_bars();
    mvprintw(0, 0, "%s", bar);
    mvprintw(0, 1, "%s", tag);
    mvprintw((row-2), 0, "%s", bar);
    (decoy == 1) ? memset(&flag[0], 'D', 1) : memset(&flag[0], ' ', 1);
    (logs == 1) ? memset(&flag[1], 'L', 1) : memset(&flag[1], ' ', 1);
    (voice == 1) ? memset(&flag[2], 'V', 1) : memset(&flag[2], ' ', 1);
    (sym == 1) ? memset(&flag[3], 'S', 1) : memset(&flag[3], ' ', 1);
    (away == 1) ? memset(&flag[4], 'A', 1) : memset(&flag[4], ' ', 1);
    (tor == 1) ? memset(&flag[5], 'T', 1) : memset(&flag[5], ' ', 1);
    (flood == 1) ? memset(&flag[6], 'F', 1) : memset(&flag[6], ' ', 1);
    (lazy == 1) ? memset(&flag[7], 'Z', 1) : memset(&flag[7], ' ', 1);
    (strict == 1) ? memset(&flag[8], 'X', 1) : memset(&flag[8], ' ', 1);
    (camuflage == 1) ? memset(&flag[9], 'C', 1) : memset(&flag[9], ' ', 1);
    (grp == 1) ? memset(&flag[10], 'G', 1) : memset(&flag[10], ' ', 1);
    mvprintw((row-2),0, " [%02d:%02d] [%s] remote is [%s] [%s] ",
    tm.tm_hour, tm.tm_min, nick, remo, flag);
    mvprintw((row-2), col-16, "%s", zzz);
    if(scrollback == 1) {

        memset(zzz, 0, sizeof(zzz));
        if(strcmp("vt100", trm) == 0 || strcmp("vt220", trm ) == 0) c_oldtrm();
        else c_bars();
        mvprintw((row-2), col-7, "-more-");
    }
    if(flood == 1 && (nmex >= burst)) {

        memset(zzz, 0, sizeof(zzz));
        mvprintw((row-2), col-8, "-flood-");
    }
    c_oldtrm_off();
    c_prompt();
    if(dre == 1) {

        move(y, x); dre = 0;
    }
    else {

        mvprintw((row-1), 0, " =--> ");
        move((row-1), 6);
    }
    if(scrollback == 0) {

        wrefresh(txt); wrefresh(mainwin);
    }
}

int killtor(void) {

    FILE *p = NULL;
    if((p = popen("ps aux | grep \"tor \" | head -n1 | \
    awk '{print $1}'", "r")) != NULL) {

        char u[20] = {"\0"};
        fgets(u, 20, p); u[strcspn(u, "\n")] = '\0';
        pclose(p);
        char *me = getlogin();

        if(strcmp(u,me) != 0) {

            colorerror(); wprintw(txt,
            "can't kill Tor (process owner: %s, me: %s)\n",u,me);
            return(1);
        }
        else {

            system("pkill -x tor"); tor = 0; tpid = 0; return(0);
        }
    }
    else return(1);
}

void torpid(void) {

    FILE *p = NULL;
    if((p = popen("ps xa | grep \"tor \" | head -n1 | \
    grep -v \"grep tor\" | awk '{print $1}'", "r")) != NULL) {

        char tp[N] = {"\0"}; fgets(tp, N, p);
        tpid = strtoul(tp, NULL, N); pclose(p);
    }
    else pclose(p);
    tor = (tpid != 0) ? 1 : 0;
}

char *hidden(void) {

    FILE *f = fopen(tchost, "r");
    if(f == NULL) return NULL;
    else {

        char *hostname = (char *) calloc(1, 64);
        if(hostname == NULL) { free(hostname); fclose(f); return NULL; }
        fgets(hostname,63,f);
        fclose(f); return hostname;
    }
}

void fireuptor(void) {

    char buf[STR] = {"\0"};
    strlcat(buf, "tor -f ", sizeof(buf));
    strlcat(buf, tcrc, sizeof(buf));
    strlcat(buf, " >/dev/null 2>&1", sizeof(buf));
    system(buf); sleep(5); torpid();
}

void printkauth(void) {

    hm(); c_info(); wprintw(txt,"<*> keys from %s:",tcauthdir); c_info_off();
    for(int i = 0; i < nauth; i++) {

        if(i%2 == 0) {

            wprintw(txt,"\n"); hm();
        }
        wprintw(txt,"  %d.[ ", i);
        if(strstr(aauth[i].auth, ".auth_private") != NULL) c_whit();
        wprintw(txt,"%-28s",aauth[i].auth); c_whit_off();
        wprintw(txt," ]");
    }
    wprintw(txt,"\n"); hm(); wprintw(txt,"(total count: %d)\n",nauth);
}

void printkprv(void) {

    hm(); c_info(); wprintw(txt,"<*> keys from %s:",tcauthclidir); c_whit_off();
    for(int i = 0; i < nprv; i++) {

        if(i%2 == 0) {

            wprintw(txt,"\n"); hm();
        }
        wprintw(txt,"  %d.[ ", i);
        if(strstr(aprv[i].prv, ".auth") != NULL &&
           !strstr(aprv[i].prv, ".auth_private")) c_whit();
        wprintw(txt,"%-28s",aprv[i].prv); c_whit_off();
        wprintw(txt," ]");
    }
    wprintw(txt,"\n"); hm(); wprintw(txt,"(total count: %d)\n",nprv);
}

void sketchyconf(FILE *f) {

    if(feof(f)) {

        endwin();
        errx(1, "sketchy tc.conf incomplete or corrupted");
    }
}

int checkfsize(FILE *fp, char *f, long int z, int cli) {

    if(fp == NULL) return(1);
    fseek(fp, 0L, SEEK_END);
    long int si = ftell(fp);
    fclose(fp);
    if(((si > z) || (si <= 0)) && cli == 0) {

        colorerror();
        wprintw(txt, "%s empty or greater than %ld byte. (%ld byte)\n",f,z,si);
        return(1);
    }
    if(((si > z) || (si <= 0)) && cli == 1) {

        errx(1, "%s empty or greater than %ld byte. (%ld byte)",f,z,si);
    }
    return(0);
}

int loadconfig(void) {

    FILE *f = fopen(tcconf, "r");
    char *b = NULL;
    size_t len = 0;
    char *line = b;
    getline(&b, &len, f);
    b[strcspn(b, "\n")] = '\0';
    if(strcmp(b,"# passphrase") != 0)
        errx(1,"irregular tc.conf (# passphrase)");
    sketchyconf(f);
    getline(&b, &len, f); b[strcspn(b, "\n")] = '\0';
    sketchyconf(f);
    strlcpy(PASS, b, sizeof(PASS));
    getline(&b, &len, f);
    b[strcspn(b, "\n")] = '\0';
    if(strcmp(b,"# Tor proxy address and port") != 0)
        errx(1,"irregular tc.conf (# Tor proxy address and port)");
    sketchyconf(f);
    getline(&b, &len, f); b[strcspn(b, "\n")] = '\0';
    sketchyconf(f);
    strlcpy(TPROXY, strsep(&b, ":"), sizeof(TPROXY));
    strlcpy(tport, strsep(&b, ":"), sizeof(tport));  /* x socks4astr()*/
    TPORT = atoi(tport);
    getline(&b, &len, f);
    b[strcspn(b, "\n")] = '\0';
    if(strcmp(b,"# tc address and port") != 0)
        errx(1,"irregular tc.conf (# tc address and port)");
    sketchyconf(f);
    getline(&b, &len, f); b[strcspn(b, "\n")] = '\0';
    sketchyconf(f);
    strlcpy(TCDAEMON, strsep(&b, ":"), sizeof(TCDAEMON));
    TCPORT = atoi(strsep(&b, ":"));
    b = NULL;
    len = 0;
    getline(&b, &len, f);
    b[strcspn(b, "\n")] = '\0';
    if(strcmp(b,"# colors") != 0) errx(1,"irregular tc.conf (# colors)");
    sketchyconf(f);
    getline(&b, &len, f); line = b; info_f = atoi(strsep(&line, ":"));
    sketchyconf(f);
    info_b = atoi(strsep(&line, ":"));
    getline(&b, &len, f); line = b; bars_f = atoi(strsep(&line, ":"));
    sketchyconf(f);
    bars_b = atoi(strsep(&line, ":"));
    getline(&b, &len, f); line = b; text_f = atoi(strsep(&line, ":"));
    sketchyconf(f);
    text_b = atoi(strsep(&line, ":"));
    getline(&b, &len, f); line = b; erro_f = atoi(strsep(&line, ":"));
    sketchyconf(f);
    erro_b = atoi(strsep(&line, ":"));
    b = NULL;
    len = 0;
    int i = 0;
    getline(&b, &len, f);
    b[strcspn(b, "\n")] = '\0';
    if(strcmp(b,"# contacts") != 0) errx(1,"irregular tc.conf (# contacts)");
    sketchyconf(f);
    while ((getline(&b,&len,f) != -1) || (strcmp(b,"#")) == 0) {

        sketchyconf(f);
        line = b;
        if(line == NULL) break;
        if((strcmp(line, "\n") == 0)) break;
        if(strstr(line, ":") == NULL) break;
        strlcpy(c[i].name, strsep(&line, ":"), sizeof(c[i].name));
        strlcpy(c[i].key, strsep(&line, ":"), sizeof(c[i].key));
        strlcpy(c[i].onion, strsep(&line, ":"), sizeof(c[i].onion));
        i++;
        nusr++;
        if(i >= MAX_CONTACTS) break;
    }
    b[strcspn(b, "\n")] = '\0';
    if(strcmp(b,"# groups") != 0) errx(1,"irregular tc.conf (# groups)");
    sketchyconf(f);
    b = NULL;
    len = 0;
    i = 0;
    while (getline(&b,&len,f) != -1) {

        sketchyconf(f);
        line = b;
        if(line == NULL) break;
        if((strcmp(line, "\n") == 0)) break;
        if(strstr(line, "# eof") != NULL) break;
        strlcpy(g[i].grpname, strsep(&line, ":"),
        sizeof(g[i].grpname));

        for(int j = 0; j < MAX_MEMBERS; j++) {

            if(line == NULL) break;
            if(strstr(line, ",") == NULL) break;
            strlcpy(g[i].members[j], strsep(&line, ","),
            sizeof(g[i].members[j]));
        }
        i++;
        ngrp++;
        if(i >= MAX_GROUPS) break;
    }
    free(b); fclose(f); return(0);
}

void gettorver(void) {

    FILE *tv = NULL;
    if((tv = popen("tor --version | head -n1 | cut -c 13-20","r")) == NULL)
        pclose(tv);
    else {

        fgets(torver, N, tv);
        torver[strcspn(torver, "\n")] = '\0';
        pclose(tv);
    }
}

void getgnupgver(void) {

    FILE *gv = NULL;
    if((gv = popen("gpg --version | head -n1 | cut -c 13-17","r")) == NULL)
        pclose(gv);
    else {

        fgets(gpgver, N, gv);
        gpgver[strcspn(gpgver, "\n")] = '\0';
        pclose(gv);
    }
}

void loadkeysid(void) {

    /* since v2.4.1 gnupg signature list show [self-signature] not uid */
    FILE *p = NULL;
    int finger = 0;
    if((p = popen("gpg --no-option --list-sigs | \
    grep \"sig 3\" | cut -c 14-29", "r")) == NULL) pclose(p);
    else {

        while(p != NULL) {

            fgets(ring[nkey].descr, 60, p);
            strlcpy(ring[nkey].keyid, ring[nkey].descr, 17);
            if(ring[nkey].keyid[0] != '\n') {

                if(ring[nkey].keyid[0] == '\0') break;
                ring[nkey].keyid[strcspn(ring[nkey].keyid, "\n")] = '\0';
                nkey++;
            }
            else ring[nkey].keyid[strcspn(ring[nkey].keyid, "\n")] = '\0';

            if(nkey >= MAX_CONTACTS) break;
        }
        pclose(p);
    }
    p = NULL;
    if((p = popen("gpg --no-option --list-sigs | \
    grep \"uid\" | cut -c 25-80", "r")) == NULL) pclose(p);
    else {

        int nk = 0;
        char buf[65] = {"\0"};
        while(p != NULL) {

            fgets(ring[nk].descr, 65, p);
            strlcpy(buf, ring[nk].keyid, sizeof(buf));
            strlcat(buf, ring[nk].descr, sizeof(buf));
            strlcpy(ring[nk].descr, buf, 65);
            if(ring[nk].descr[0] != '\n') {

                if(ring[nk].descr[0] == '\0') break;
                ring[nk].descr[strcspn(ring[nk].descr, "\n")] = '\0';
                nk++;
            }
            else ring[nk].descr[strcspn(ring[nk].descr, "\n")] = '\0';

            if(nk >= MAX_CONTACTS) break;
        }
        pclose(p);
    }
    p = NULL; /* ed25519 fingerprint */
    if((p = popen("gpg --no-option --list-sigs |\
    grep -A1 \"pub   ed25519\" | grep -v \"pub   ed25519\" |\
    grep -v -e \"pub   ed25519\" -e \"--\" |\
    cut -c 7-46" , "r")) == NULL) pclose(p);
    else {

        while(p != NULL) {

            fgets(ring[nkey].descr, 60, p);
            strlcpy(ring[nkey].keyid, ring[nkey].descr, 41);
            if(ring[nkey].keyid[0] != '\n') {

                if(ring[nkey].keyid[0] == '\0') break;
                ring[nkey].keyid[strcspn(ring[nkey].keyid, "\n")] = '\0';
                nkey++;
                finger++;
            }
            else ring[nkey].keyid[strcspn(ring[nkey].keyid, "\n")] = '\0';

            if(nkey >= MAX_CONTACTS) break;
        }
        pclose(p);
    }
    p = NULL;  /* ed25519 description */
    if((p = popen("gpg --no-option --list-sigs | \
    grep -A3 \"ed25519\" | grep -v -e \"sig\" -e \"pub\" -e \"sub\" -e \"--\"|\
    grep \"uid\" | cut -c 25-90", "r")) == NULL) pclose(p);
    else {

        int nk = (nkey - finger);
        char buf[65] = {"\0"};
        while(p != NULL) {

            fgets(ring[nk].descr, 64, p);
            strlcpy(buf, ring[nk].keyid, sizeof(buf));
            strlcat(buf, ring[nk].descr, sizeof(buf));
            strlcpy(ring[nk].descr, buf, 75);
            if(ring[nk].descr[0] != '\n') {

                if(ring[nk].descr[0] == '\0') break;
                ring[nk].descr[strcspn(ring[nk].descr, "\n")] = '\0';
                nk++;
            }
            else ring[nk].descr[strcspn(ring[nk].descr, "\n")] = '\0';

            if(nk >= MAX_CONTACTS) break;
        }
        pclose(p);
    }
}

void loadauthkeys(void) {

    memset(&aauth, 0, sizeof(aauth)); /* clear struct */
    nauth = 0;
    char buf[STR] = {"\0"};
    char tmp[29] = {"\0"};
    strlcat(buf, "ls -1 ", sizeof(buf));
    strlcat(buf, tcauthdir, sizeof(buf));
    FILE *p = popen(buf, "r");
    if(p == NULL) pclose(p);
    else {

        while(p != NULL) { /* load .auth files */

        tmp[0] = '\0';
        fgets(tmp,29,p);
        int l = strlen(tmp);
        if( l == 0 || l > 28 ) break;
        strlcpy(aauth[nauth].auth,tmp,sizeof(aauth[nauth].auth));
        aauth[nauth].auth[strcspn(aauth[nauth].auth,"\n")] = '\0';
        nauth++;

        if(nauth >= N) break;
        }
        pclose(p);
    }
    memset(&aprv, 0, sizeof(aprv));
    nprv = 0; buf[0] = '\0'; tmp[0] = '\0';
    strlcat(buf, "ls -1 ", sizeof(buf));
    strlcat(buf, tcauthclidir, sizeof(buf));
    p = popen(buf, "r");
    if(p == NULL) pclose(p);
    else {

        while(p != NULL) { /* load .auth_private files */

        tmp[0] = '\0';
        fgets(tmp,29,p);
        int l = strlen(tmp);
        if( l == 0 || l > 28 ) break;
        strlcpy(aprv[nprv].prv,tmp,sizeof(aprv[nprv].prv));
        aprv[nprv].prv[strcspn(aprv[nprv].prv,"\n")] = '\0';
        nprv++;

        if(nprv >= N) break;
        }
        pclose(p);
    }
}

void onoff(int i, int onoff, int a) {

    char buf[STR] = {"\0"}; char f[STR] = {'\0'};
    if(a == 0) { /* client side auth for remote access */

        strlcat(buf, tcauthdir, sizeof(buf));
        strlcat(buf, "/", sizeof(buf));
        strlcat(buf, aauth[i].auth, sizeof(buf));
    }
    if(a == 1) { /* server side auth for local access */

        strlcat(buf, tcauthclidir, sizeof(buf));
        strlcat(buf, "/", sizeof(buf));
        strlcat(buf, aprv[i].prv, sizeof(buf));
    }
    strlcat(f, buf, sizeof(f));
    const unsigned int l = strlen(f);
    if(onoff == 1 && l > 0 && f[l-1] == '@') { /* enable files */

            f[l-1] = '\0'; rename(buf, f); strict = 1;
    }
    if(onoff == 0 && l > 0 && f[l-1] != '@') { /* disable files */

            strlcat(f, "@", sizeof(f)); rename(buf, f); strict = 0;
    }
}

void mutual(int m) {

    loadauthkeys();
    if(m == 1) { /* enable all keys */

        for(int i = 0; i < nauth; i++)
        if(strstr(aauth[i].auth,".auth_private@") != NULL) onoff(i,1,0);
        for(int i = 0; i < nprv; i++)
        if(strstr(aprv[i].prv,".auth@") != NULL) onoff(i,1,1);
    }
    if(m == 0) { /* disable all keys */

        for(int i = 0; i < nauth; i++)
        if(strstr(aauth[i].auth,".auth_private") != NULL) onoff(i,0,0);
        for(int i = 0; i < nprv; i++)
        if(strstr(aprv[i].prv,".auth") != NULL) onoff(i,0,1);
    }
}

void selectk(int kk) {

    loadauthkeys(); (kk == 0) ? printkauth() : printkprv();
    int j = 0, n = 0;
    do {

        draw();
        move(row-1, 0); clrtoeol();
        mvprintw((row-1),0, " =--> [*] selected {%d} or (d)one: ", n);
        move(row-1, 34);
        echo();
        getnstr(msg, 1);
        noecho();
        if(msg[0] == '\0' || msg[0] == 'd') continue;
        int w = (atoi(msg));
        if(kk == 0 && w >= 0 && w < nauth+1) {

            if(strstr(aauth[w].auth,".auth_private@") != NULL) onoff(w,1,0);
            else if(strstr(aauth[w].auth,".auth_private") != NULL) onoff(w,0,0);
        }
        if(kk == 1 && w >= 0 && w < nprv+1) {

            if(strstr(aprv[w].prv,".auth@") != NULL) onoff(w,1,1);
            else if(strstr(aprv[w].prv,".auth") != NULL) onoff(w,0,1);
        }
        j++; n++;

    } while(j < N && (strcmp(msg, "d") != 0));
}

void nokey(void) {

    colorerror(); wprintw(txt,"no keys found\n"); wrefresh(txt);
}

void strictness(void) {

    loadauthkeys();
    switch((int)msg[0]) {
    case 'c': (nauth >= 1) ? selectk(0) : nokey(); break;
    case 's': (nprv >= 1) ? selectk(1) : nokey(); break;
    case 'm':
        if(nauth == 0 && nprv == 0) nokey();
        (strict == 0) ? mutual(1) : mutual(0);
    break;
    default: break;
    }
}

void cleanup(int sig) {

    hm(); c_whit();
    wprintw(txt, "----------------------------------------------------");
    wprintw(txt, "-------------------\n"); c_whit_off();
    hm(); wprintw(txt, "                         Enjoy the update from\n");
    hm(); wprintw(txt, "%s/tc.html\n", TCURL);
    hm(); wprintw(txt, "Donate because than I can get some money to buy lunc");
    wprintw(txt, "h and skateboards\n");
    hm(); wprintw(txt, "               bc1qwzj9tk9vym63ddfh9v8xt0guxw47mxm8k");
    wprintw(txt, "wjpkt\n"); hm(); c_whit();
    wprintw(txt, "----------------------------------------------------");
    wprintw(txt, "-------------------\n"); c_whit_off();
    draw(); unlink(TCS); shred(TCT); unlink(TCR); shred(TCN); unlink(tclock);

    FILE *f = fopen(tclog, "r");
    if(f != NULL) {

        fclose(f);
        mvprintw((row-1),0, " =--> [*] tc.log not empty. shred it? [y/n]: ");
        move(row-1, 45); echo(); getnstr(msg, 1); noecho(); curs_set(0);
        if(msg[0] == 'y') shred(tclog); fprintf(stderr, "\n");
    }
    syslog(LOG_INFO, "can become addictive, use with moderation");
    endwin(); exit(0);
}

void speak(void) {

    char buf[STR] = {"\0"};
    strlcat(buf, "espeak -f ", sizeof(buf));
    strlcat(buf, home, sizeof(buf));
    strlcat(buf, "/.tc/.tcn -a 150 -g 3 -p 30 -s 150 >/dev/null 2>&1",
    sizeof(buf));
    system(buf);
}

void addscrollback(char *rmsg) {

    char p[STR] = {"\0"};
    if(rmsg == NULL) {

        strftime(p, 7, "%H:%M ", &tm);
        strlcat(p, "<", sizeof(p));
        strlcat(p, nick, sizeof(p));
        strlcat(p, "> ", sizeof(p));
        strlcat(p, msg, sizeof(p));
        strlcat(p, "\n", sizeof(p));
        waddstr(pad, p);
    }
    else {

        strlcat(p, "\n", sizeof(p));
        strftime(p, 7, "%H:%M ", &tm);
        waddstr(pad, p);
        waddstr(pad, rmsg);
        waddstr(pad, "\n");
    }
    nrow++;
}

void drawremote(char *rmsg) {

    /* remote msg include nick; addscrollback() & logger() verify NULL nick */
    if(strstr(rmsg, "<notice> ") != NULL) {

        char buf[MSG] = {"\0"};
        strlcpy(buf, rmsg+13, 70);
        hm(); c_whit(); wprintw(txt, "*notice* "); c_whit_off();
        c_erro(); wprintw(txt, "%s\n", buf); c_erro_off();
        addscrollback(rmsg);
    }
    else if(strstr(rmsg, "<*> ") != NULL) {

        char buf[MSG] = {"\0"};
        strlcpy(buf, rmsg+4, 70);
        colorinfo();
        wprintw(txt, "%s\n", buf);
        addscrollback(rmsg);
    }
    else {

        c_text(); hm(); wprintw(txt, "%s\n", rmsg); c_text_off();
        addscrollback(rmsg);
    }
    if(strstr(rmsg," changed the topic to: ") != NULL) {

        char *m; m = memchr(rmsg, ':', strlen(rmsg)); topic(m+2);
    }
    if(logs == 1) logger(NULL, rmsg);
    dre = 1;
    if(scrollback == 0) draw();
    if(voice == 1) speak();
}

void sender(char *onion, char *n) {

    memset(zzz, 0, sizeof(zzz));
    int sockfd;
    struct sockaddr_in relay;
    memset(&relay, 0, sizeof(relay)); /* clear struct */
    if((sockfd = socket(AF_INET, SOCK_STREAM, 0)) == -1) exit(1);
    relay.sin_family = AF_INET;
    relay.sin_port = htons(TPORT);
    inet_pton(AF_INET, TPROXY, &relay.sin_addr);
    memset(&(relay.sin_zero), 0, sizeof(relay.sin_zero));
    struct timeval tv;
    tv.tv_sec = ALIVE;
    tv.tv_usec = 0;
    setsockopt(sockfd, SOL_SOCKET, SO_SNDTIMEO, &tv, sizeof(tv));
    if(connect(sockfd, (struct sockaddr *)&relay,
    sizeof(struct sockaddr)) == -1) {

        colorerror();
        wprintw(txt, "Tor proxy unrechable use '/circuit' to verify\n");
        nocom(); tor = 0; return;
    }
    if(lazy == 1) {

        close(sockfd); return;
    }
    /*
     *  https://www.openssh.com/txt/socks4.protocol
     *  https://www.openssh.com/txt/socks4a.protocol
     *  ver:4;cd:1;dest port:2323;0.0.0.x;x=1;user:NULL+onion
     */
    unsigned char buf[RWLENGTH];
    if(camuflage == 1) {

        /*                      ver  cd   23   23   0    0    0    1  */
        unsigned char head[8]={0x04,0x01,0x09,0x13,0x00,0x00,0x00,0x01};
        unsigned char zero[1]={0x00};
        char fakie[50] = {"\0"};
        strlcpy(fakie, mkfake(1), sizeof(fakie));
        int lenz = strlen(fakie);
        memcpy(buf, head, 8);
        memcpy(buf+8, fakie, lenz); /* socks4 username filed */
        memcpy(buf+8+lenz, zero, 1);
        memcpy(buf+8+lenz+1, onion, 63);
        send(sockfd, buf, (8+lenz+1+63), 0);
    }
    if(camuflage == 0) {

        /*                      ver  cd   23   23   0    0    0    1   user*/
        unsigned char head[9]={0x04,0x01,0x09,0x13,0x00,0x00,0x00,0x01,0x00};
        memcpy(buf, head, 9);
        memcpy(buf+9, onion, 63);
        send(sockfd, buf, 72, 0); /* 63+9 = head + onion v3 addr */
    }
    recv(sockfd, buf, 16, 0);     /* need recieve: 00 5a 00 00 00 */
    if(buf[0] || buf[1] != 90) {  /* 5a == 90 */

        if(n != NULL && strcmp("decoy",n) == 0) return;
        colorerror(); wprintw(txt, "can't send!\n");
        memcpy(zzz, onion, 6); memcpy(zzz+6, "... zZZzz", 9);
        getyx(mainwin, y, x);
        draw();
        move(y, x);
        close(sockfd);
        return;
    }
    else {

        tor = 1;
        FILE *fs = fopen(TCS, "r");
        if(fs == NULL) exit(1);
        memset(buf, 0, sizeof(buf));
        int block;
        while((block = fread(buf, sizeof(char), RWLENGTH, fs)) > 0) {

            if(send(sockfd, buf, block, 0) < 0) break;
            memset(buf, 0, sizeof(buf));
        }
        close(sockfd);
        fseek(fs, 0, SEEK_END);
        long flen = ftell(fs);
        byteout =  byteout + flen;
        fclose(fs);
    }
}

bool isbin(const void *data, size_t len) {

    return memchr(data, '\0', len) != NULL;
}

int checkfile(void) {

    if(access(msg, F_OK) != 0) {

        colorerror(); wprintw(txt, "%s not exist\n", msg); return(1);
    }
    struct stat st;
    if(stat(msg, &st) == 0) {

        if(st.st_mode & S_IFDIR) {

            colorerror(); wprintw(txt, "%s it's a directory\n", msg); return(1);
        }
        if(st.st_size == 0 && S_ISREG(st.st_mode)) {

            colorerror(); wprintw(txt, "%s it's empty\n", msg); return(1);
        }
        if(st.st_uid != getuid()) {

            colorerror();
            wprintw(txt, "%s owner uid: %d myself: %d won't open\n",
            msg, st.st_uid, getuid());
            return(1);
        }
        if((st.st_mode & (S_IRUSR|S_IRGRP|S_IROTH)) &&
           (S_ISREG(st.st_mode))) return(0);
    }
    return(0);
}

void enc_sym(int z) {

    char buf[300] = {"\0"};
    if(symg == 1) {

        if(atoi(&gpgver[0]) == 1) return; /* gnupg v.1.x is not supported */
        strlcat(buf, "gpg --trust-model always --yes -q -e -c \
        --cipher-algo ", sizeof(buf));
        strlcat(buf, cipher, sizeof(buf));
        strlcat(buf," --passphrase-repeat 0 --batch --passphrase ",sizeof(buf));
        strlcat(buf, K, sizeof(buf));
        strlcat(buf, " -R ", sizeof(buf));
        strlcat(buf, k[z], sizeof(buf));
        strlcat(buf, " -o ", sizeof(buf));
        strlcat(buf, home, sizeof(buf));
        strlcat(buf, "/.tc/.tcs ", sizeof(buf));
        strlcat(buf, home, sizeof(buf));
        strlcat(buf, "/.tc/.tct", sizeof(buf));
        strlcat(buf, " >/dev/null 2>&1", sizeof(buf));
    }
    if(symo == 1) {

        strlcat(buf, "openssl enc -aes-256-cbc -pbkdf2 -k ", sizeof(buf));
        strlcat(buf, K, sizeof(buf));
        strlcat(buf, " -in ", sizeof(buf));
        strlcat(buf, home, sizeof(buf));
        strlcat(buf, "/.tc/.tct", sizeof(buf));
        strlcat(buf, " -out ", sizeof(buf));
        strlcat(buf, home, sizeof(buf));
        strlcat(buf, "/.tc/.tcs >/dev/null 2>&1", sizeof(buf));
    }
    system(buf); shred(TCT);
}

void enc_sym_file(int z, char *file) {

    char buf[300] = {"\0"};
    if(symg == 1) {

        if(atoi(&gpgver[0]) == 1) return; /* gnupg v.1.x is not supported */
        strlcat(buf, "gpg --trust-model always --yes -q -e -c \
        --cipher-algo ", sizeof(buf));
        strlcat(buf, cipher, sizeof(buf));
        strlcat(buf," --passphrase-repeat 0 --batch --passphrase ",sizeof(buf));
        strlcat(buf, K, sizeof(buf));
        strlcat(buf, " -R ", sizeof(buf));
        strlcat(buf, k[z], sizeof(buf));
        strlcat(buf, " -o ", sizeof(buf));
        strlcat(buf, home, sizeof(buf));
        strlcat(buf, "/.tc/.tcs ", sizeof(buf));
        strlcat(buf, file, sizeof(buf));
        strlcat(buf, " >/dev/null 2>&1", sizeof(buf));
    }
    if(symo == 1) {

        strlcat(buf, "openssl enc -aes-256-cbc -pbkdf2 -k ", sizeof(buf));
        strlcat(buf, K, sizeof(buf));
        strlcat(buf, " -in ", sizeof(buf));
        strlcat(buf, file, sizeof(buf));
        strlcat(buf, " -out ", sizeof(buf));
        strlcat(buf, home, sizeof(buf));
        strlcat(buf, "/.tc/.tcs >/dev/null 2>&1", sizeof(buf));
    }
    system(buf); shred(TCT);
}

void enc_asym(int z) {

    if(atoi(&gpgver[0]) == 1) return; /* gnupg v.1.x is not supported */
    char buf[STR] = {"\0"};
    strlcat(buf, "gpg --trust-model always --yes -q -e -R ",sizeof(buf));
    strlcat(buf, k[z], sizeof(buf));
    strlcat(buf, " -o ", sizeof(buf));
    strlcat(buf, home, sizeof(buf));
    strlcat(buf, "/.tc/.tcs ", sizeof(buf));
    strlcat(buf, home, sizeof(buf));
    strlcat(buf, "/.tc/.tct >/dev/null 2>&1", sizeof(buf));
    system(buf); shred(TCT);
}

void enc_asym_file(int z, char *file) {

    if(atoi(&gpgver[0]) == 1) return; /* gnupg v.1.x is not supported */
    char buf[STR] = {"\0"};
    strlcat(buf, "gpg --trust-model always --yes -q -e -R ",sizeof(buf));
    strlcat(buf, k[z], sizeof(buf));
    strlcat(buf, " -o ", sizeof(buf));
    strlcat(buf, home, sizeof(buf));
    strlcat(buf, "/.tc/.tcs ", sizeof(buf));
    strlcat(buf, file, sizeof(buf));
    strlcat(buf, " >/dev/null 2>&1", sizeof(buf));
    system(buf); shred(TCT);
}

void ucrypt(char *n, char *ms, char *file) {

    pthread_mutex_lock(&mtxtct);
    if(file == NULL && n != NULL && ms != NULL) { /* NULL file means msg */

        FILE *f = fopen(TCT, "w");
        fprintf(f, "<%s> %s\n", n, ms);
        fclose(f);
        if(sym == 1) enc_sym(0);
        if(sym == 0) enc_asym(0);
        sender(o[0], n);
    }
    else {

        if(sym == 1) enc_sym_file(0,file);
        if(sym == 0) enc_asym_file(0,file);
        sender(o[0], n);
    }
    if(pthread_mutex_unlock(&mtxtct) != 0) wprintw(txt,"mutex unlock failed\n");
}

void gcrypt(char *n, char *mg, char *file) {

    pthread_mutex_lock(&mtxtct);
    if(file == NULL && n != NULL && mg != NULL) {

        FILE *f = fopen(TCT, "w");
        fprintf(f, "<%s> %s\n", n, mg);
        fclose(f);
        if(sym == 1) {

            enc_sym(0);
            for(int i = 0; i < who; i++) sender(o[i], n);
        }
        if(sym == 0) {

            for(int i = 0; i < who; i++) {

                enc_asym(i); sender(o[i], n);
            }
        }
    }
    else {

        if(sym == 1) {

            enc_sym_file(0,file);
            for(int i = 0; i < who; i++) sender(o[i], n);
        }
        if(sym == 0) {

            for(int i = 0; i < who; i++) {

                enc_asym_file(i,file); sender(o[i], n);
            }
        }
    }
    if(pthread_mutex_unlock(&mtxtct) != 0) wprintw(txt,"mutex unlock failed\n");
}

void forward(void) {

    (grp == 1) ? gcrypt("*", msg, NULL) : ucrypt("*", msg, NULL);
}

void sendinfos(char *ss) {

    char buf[STR] = {"\0"};
    strlcat(buf, nick, sizeof(buf));
    strlcat(buf, ss, sizeof(buf));
    strlcpy(msg, buf, sizeof(msg));
    forward();
}

void algoz(void) {

    hm(); c_info(); wprintw(txt, "<*> symmetric cipher algos:"); c_info_off();
    for(int i = 0; i < (SKA-1); i++) {

        if(i%3 == 0) {

            wprintw(txt,"\n"); hm();
        }
        wprintw(txt,"  %d.[ ", i);
        if(strcmp(cipher, alg[i]) == 0) c_whit();
        wprintw(txt, "%-11s", alg[i]); c_whit_off();
        wprintw(txt," ]");
    }
    wprintw(txt, "\n"); hm(); wprintw(txt, "(total count: %d)\n", (SKA-1));
    draw();

    if(con == 1 && sym == 1 && symg == 1) {

        mvprintw((row-1),0, " =--> [*] choose cipher: ");
        move(row-1, 25);
        echo();
        getnstr(msg, 1);
        noecho();
        curs_set(0);
        if(msg[0] == '\0') return;
        if(!isdigit(msg[0])) return;
        int on = (atoi(msg));
        if(on >= 0 && on < (SKA-1)) {

            memset(cipher, 0, sizeof(cipher));
            strlcpy(cipher, alg[on], sizeof(cipher));
            colorinfo();
            wprintw(txt, "encryption cipher switched to %s\n", cipher);
            draw();
            char buf[MSG] = {"\0"};
            strlcat(buf, " has switched cipher to: ", sizeof(buf));
            strlcat(buf, cipher, sizeof(buf));
            sendinfos(buf);
        }
    }
}

void decrypt(int sig) {

    char buf[STR] = {"\0"};
    if(sym == 1 && symo == 1) {

        strlcat(buf, "openssl enc -aes-256-cbc -pbkdf2 -k ",sizeof(buf));
        strlcat(buf, K, sizeof(buf));
        strlcat(buf, " -in ", sizeof(buf));
        strlcat(buf, home, sizeof(buf));
        strlcat(buf, "/.tc/.tcr", sizeof(buf));
        strlcat(buf, " -out ", sizeof(buf));
        strlcat(buf, home, sizeof(buf));
        strlcat(buf, "/.tc/.tcn -d >/dev/null 2>&1", sizeof(buf));
    }
    if(sym == 1 && symg == 1) {

        strlcat(buf,
        "gpg --pinentry-mode loopback --yes -q --batch --passphrase ",
        sizeof(buf));
        strlcat(buf, K, sizeof(buf));
        strlcat(buf, " -o ", sizeof(buf));
        strlcat(buf, home, sizeof(buf));
        strlcat(buf, "/.tc/.tcn ", sizeof(buf));
        strlcat(buf, " -d ", sizeof(buf));
        strlcat(buf, home, sizeof(buf));
        strlcat(buf, "/.tc/.tcr >/dev/null 2>&1", sizeof(buf));
    }
    if(sym == 0 && symg == 0 && symo == 0) {

        strlcat(buf,
        "gpg --pinentry-mode loopback --yes -q --batch --passphrase ",
        sizeof(buf));
        strlcat(buf, PASS, sizeof(buf));
        strlcat(buf, " -o ", sizeof(buf));
        strlcat(buf, home, sizeof(buf));
        strlcat(buf, "/.tc/.tcn ", sizeof(buf));
        strlcat(buf, home, sizeof(buf));
        strlcat(buf, "/.tc/.tcr >/dev/null 2>&1", sizeof(buf));
    }
    if(access(TCR, F_OK) == 0) system(buf);
    unlink(TCR);
    FILE *f = fopen(TCN, "r");
    if(f == NULL) return;
    char rem[MSG] = {"\0"};
    if(fgets(rem, MSG, f) == NULL) *rem = '\0';
    rem[strcspn(rem, "\n")] = '\0';
    if(strstr(rem, "<decoy>") != NULL) {

        fclose(f); shred(TCN); return;
    }
    fseek(f, 0L, SEEK_END);
    long int siz = ftell(f);
    if(siz > MSG) { /* assume as file */

        char fn[STR];
        int bn;
        strlcpy(fn, tcrcvfile, sizeof(fn));
        if((bn = mkstemp(fn)) == -1) { unlink(fn); close(bn); }
        rename(TCN, fn);
        fclose(f);
        colorinfo();
        wprintw(txt, "file recieved: %s (%ld bytes)\n", fn, siz);
        draw();
        char b[STR] = {"\0"};
        char l[20] = {"\0"};
        strlcat(b, "<*> file recieved: ", sizeof(b));
        strlcat(b, fn, sizeof(b));
        strlcat(b, " (", sizeof(b));
        snprintf(l, 20, "%ld", siz);
        strlcat(b, l, sizeof(b));
        strlcat(b, " bytes)", sizeof(b));
        addscrollback(b);
        if(logs == 1) logger(NULL, b);
        shred(TCN);
    }
    else { /* its a msg; tiny files are displayed in-line */

        getyx(mainwin, y, x);
        if(rem[0] != '\0') drawremote(rem);
        fclose(f);
        shred(TCN);
    }
    /* notify away if joined and not idle, avoid ping-pong away notice */
    if(con == 1 && away == 1 && grp == 0) {

        if(strstr(rem, " is away: ") != NULL) return;
        else {

            char re[MSG] = {"\0"};
            strlcat(re, " is away: ", sizeof(re));
            strlcat(re, way, sizeof(re));
            sendinfos(re);
        }
    }
    /* prevent blank line in symmetric if someone tcp CONNECT to tc daemon */
    if(strstr(rem, " activated symmetric ") != NULL)
        noblank = 1;
    if(strstr(rem, " has switched OFF symmetric encyption") != NULL)
        noblank = 0;
    return;
}

void showcontacts(char *nam) {

    if(nam == NULL) {

        hm(); c_info(); wprintw(txt, "<*> contacts:"); c_info_off();
        for(int i = 0; i < nusr; i++) {

            if(i%6 == 0) {

                wprintw(txt, "\n"); hm(); wprintw(txt, " ");
            }
            wprintw(txt, "[");
            c_whit(); wprintw(txt, "%-9s", c[i].name); c_whit_off();
            wprintw(txt, "] ");
        }
        wprintw(txt, "\n"); hm(); wprintw(txt, "(total count: %d)\n", nusr);
    }
    else {

        int found = 0;
        for(int i = 0; i < nusr; i++) {

            if(strcmp(c[i].name, nam) == 0) {

                hm();
                c_info(); wprintw(txt, "<*> %s:\n", c[i].name); c_info_off();
                hm(); wprintw(txt, "keyid: ");
                c_whit(); wprintw(txt, "%s\n", c[i].key); c_whit_off();
                hm(); wprintw(txt, "onion: ");
                c_erro(); wprintw(txt, "%s\n", c[i].onion); c_erro_off();
                found = 1;
            }
        }
        if(found == 0) {

            colorerror(); wprintw(txt, "%s not found in contacts\n", nam);
        }
    }
}

void showgroups(void) {

    hm(); c_info(); wprintw(txt, "<*> groups:"); c_info_off();
    for(int i = 0; i < ngrp; i++) {

        wprintw(txt, "\n");
        hm(); c_whit(); wprintw(txt, "%-13s", g[i].grpname); c_whit_off();
        for(int q = 0; q < MAX_MEMBERS; q++) {

            wprintw(txt, "[");
            c_erro(); wprintw(txt, "%-9s", (char *)&g[i].members[q]);
            c_erro_off(); wprintw(txt, "] ");
        }
    }
    wprintw(txt, "\n"); hm(); wprintw(txt, "(total group: %d)\n", ngrp);
}

void history(char *mex) {

    if(mex == NULL) {

        hm(); c_info(); wprintw(txt, "<*> history:\n"); c_info_off();
        for(int i = 0; i < hpos; i++) {

            hm(); wprintw(txt, "[%d] %s\n", i, s[i]);
        }
    }
    else {

        strlcpy(s[hpos], mex, MSG); hpos++; if(hpos == MSG) hpos = 0;
    }
}

int checkkey(char *nam, char *ck) {

    int keyok = 0;
    for(int i = 0; i < nkey; i++) {

        if(ck == NULL) break; if(strcmp(ring[i].keyid, ck) == 0) keyok = 1;
    }
    if(keyok == 1) return(0);
    else {

        colorerror();
        wprintw(txt, "%s keyid %s not found in local PGP key ring\n", nam, ck);
        return(1);
    }
}

int loadu(char *n) {

    int keynok = 0, found = 0;
    for(int i = 0; i < nusr; i++) {

        if(strcmp(c[i].name, n) == 0) {

            if(checkkey(c[i].name, c[i].key) == 0) {

                found = 1;
                k[who] = c[i].key;
                o[who] = c[i].onion;
                strlcpy(remo, n, sizeof(remo));
                who++;
            }
            else keynok = 1;
        }
    }
    if(found == 0) {

        colorerror(); wprintw(txt, "%s not found in contacts\n", n);
    }
    if(keynok == 1 || found == 0) return(1);
    return(0);
}

int loadg(char *n) {

    for(int i = 0; i < ngrp; i++) {

        if(strcmp(g[i].grpname, n) == 0) {

            for(int q = 0; q < MAX_MEMBERS; q++) {

                if(strcmp(g[i].members[q], "\n") == 0) break;
                if(strcmp(g[i].members[q], "") == 0) break;
                if(loadu(g[i].members[q]) == 0) {

                    grp = 1; strlcpy(remo, g[i].grpname, sizeof(remo));
                }
            }
        }
    }
    if(grp == 0) return(1);
    return(0);
}

int checkname(char *n) {

    int uu = 0, gg = 0;
    for(int j = 0; j <= nusr+1; j++) { if(strcmp(c[j].name, n) == 0) uu++; }
    for(int j = 0; j <= ngrp+1; j++) { if(strcmp(g[j].grpname, n) == 0) gg++; }
    if(uu == 0 && gg == 0) {

        colorerror(); wprintw(txt,"%s not found in contacts\n", n); return(1);
    }
    if(uu > 1 ) {

        colorerror();
        wprintw(txt, "too many entry found for \"%s\" in contacts\n", n);
        return(1);
    }
    if(gg > 1 ) {

        colorerror();
        wprintw(txt, "too many entry found for \"%s\" in groups\n", n);
        return(1);
    }
    if(uu == 1 && gg == 1) {

        move(row-1, 0);
        clrtoeol();
        mvprintw((row-1),0,
        " =--> [*] too many entry; is (c)ontact or (g)roup? [c/g]: ");
        move(row-1, 58);
        echo();
        char e[2] = {"\0"};
        getnstr(e, 1);
        noecho();
        move(row-1, 0);
        clrtoeol();

        if(e[0] == 'c')      { if(loadu(n) == 1) return(1); }
        else if(e[0] == 'g') { if(loadg(n) == 1) return(1); }
        else return(1);
    }
    if(uu == 1 && gg == 0)   { if(loadu(n) == 1) return(1); }
    if(uu == 0 && gg == 1)   { if(loadg(n) == 1) return(1); }
    return(0);
}

int saveonion(void) {

    if(hidden() != NULL) {

        char current[64] = {"\0"};
        strlcat(current, hidden(), sizeof(current));
        hm(); wprintw(txt, "current: %s\n", current);
        if(killtor() == 1) return(1);
        nocom(); draw();
        char buf[STR] = {"\0"};
        strlcat(buf, "mv ", sizeof(buf));
        strlcat(buf, home, sizeof(buf));
        strlcat(buf, "/.tc/tor ", sizeof(buf));
        strlcat(buf, home, sizeof(buf));
        strlcat(buf, "/.tc/tor.", sizeof(buf));
        strlcat(buf, current, sizeof(buf));
        system(buf);
    }
    return(0);
}

void socks4astr(void) {

    socks4a[0] = '\0';
    strlcat(socks4a, "curl -m 30 --retry 2 -k -s -L -x socks4a://",
    sizeof(socks4a));
    strlcat(socks4a, TPROXY, sizeof(socks4a));
    strlcat(socks4a, ":", sizeof(socks4a));
    strlcat(socks4a, tport, sizeof(socks4a));
    strlcat(socks4a, " ", sizeof(socks4a));
}

void alive(void) {

    memset(zzz, 0, sizeof(zzz));
    getyx(mainwin, y, x);
    if(strcmp("vt100", trm) == 0 || strcmp("vt220", trm ) == 0) c_oldtrm();
    else c_bars();
    mvprintw((row-2), col-6, "-tor-"); c_oldtrm_off(); wrefresh(mainwin);
    move(y, x);
    FILE *p = popen("curl -m 30 -k -s -L --noproxy torproject.org \
    check.torproject.org | grep \"Your IP address appears to be\" | \
    cut -d \\> -f3 | cut -d \\< -f1", "r");
    if(p == NULL) pclose(p);
    else {

        fgets(chk, 17, p);
        chk[strcspn(chk, "\n")] = '\0';
        pclose(p);
    }
    fireuptor();
    char buf[STR] = {"\0"};
    strlcat(buf, socks4a, sizeof(buf));
    strlcat(buf, "check.torproject.org | \
    grep \"Your IP address appears to be\" | \
    cut -d \\> -f3 | cut -d \\< -f1", sizeof(buf));
    p = popen(buf, "r");
    if(p == NULL) pclose(p);
    else {

        fgets(chktor, 17, p); chktor[strcspn(chktor, "\n")] = '\0'; pclose(p);
    }
    if((strcmp(&chk[0], "\0") == 0)) {

        FILE *py = popen("curl -m 30 -k -s -L --noproxy dyndns.org \
        checkip.dyndns.org | cut -d\":\" -f2 | cut -d\"<\" -f1", "r");
        if(py == NULL) pclose(py);
        else {

            fgets(chk, 17, py); chk[strcspn(chk, "\n")] = '\0'; pclose(py);
        }
    }
    if((strcmp(&chktor[0], "\0") == 0)) {

        fireuptor();
        buf[0] = '\0';
        strlcat(buf, socks4a, sizeof(buf));
        strlcat(buf, "checkip.dyndns.org | cut -d\":\" -f2 | cut -d\"<\" -f1",
        sizeof(buf));
        FILE *py = NULL;
        py = popen(buf, "r");
        if(py == NULL) pclose(py);
        else {

            fgets(chktor, 17, py); chktor[strcspn(chktor, "\n")] = '\0';
            pclose(py);
        }
    }
    c_prompt();
}

void *clocker(void *arg) {

    time(&t);
    tm = *localtime(&t);
    unsigned int sec;
    time_t up, t2 = time(&t);
    up = difftime(t2,tstart);
    sec = up % 60;
    sleep(60 - sec);

    while (1) {

        time(&t);
        tm = *localtime(&t);
        getyx(mainwin, y, x);
        draw();
        move(y, x);
        refresh();
        if(tm.tm_hour == 0 && tm.tm_min == 0) {

            colorinfo(); wprintw(txt, "day changed to %s", ctime(&t)); draw();
        }
        sleep(60);
        time(&t);
        tm = *localtime(&t);
        getyx(mainwin, y, x);
        draw();
        move(y, x);
        refresh();
        interval++;
        if(flood == 1 && (nmex > burst)) nmex = 0; /* 1 min  */
        if((interval % 2) == 0 && con == 1 && flood == 1) nmex = 0;
        if((interval % ALIVE) == 0 && con == 1) {

            alive(); interval = 0;
        }
        if((interval % ALIVE) == 0 && con == 0) interval = 0;
    }
}

void *send_decoy(void *arg) {

    while(con == 1) {

        int i = 1;
        int deception = lrand48() % 120;
        if(ndec > 8) { /* burst minimum is 10 */

            sleep(90); ndec = 0; continue;
        }
        sleep(deception);
        int fake = lrand48() % 7; /* minor of MSG */
        char buf[MSG] = {"\0"};
        while (i <= fake) {

            strlcat(buf, mkfake(0), sizeof(buf)); i++;
        }
        (grp == 1) ? gcrypt("decoy",buf,NULL) : ucrypt("decoy",buf,NULL);
        ndec++;
    }
    return NULL;
}

void *listener(void *arg) {

    int sockfd = 0;
    int nsockfd = 0;
    struct sockaddr_in local;   /* tc server */
    struct sockaddr_in remote;  /* client address */
    memset(&local, 0, sizeof(local)); /* clear struct */
    memset(&remote, 0, sizeof(remote)); /* clear struct */
    char revbuf[RWLENGTH];
    if((sockfd = socket(AF_INET, SOCK_STREAM, 0)) == -1) exit(1);
    local.sin_family = AF_INET;
    local.sin_port = htons(TCPORT);
    inet_pton(AF_INET, TCDAEMON, &local.sin_addr);
    memset(&local.sin_zero, 0, sizeof(local.sin_zero)); /* clear struct */
    int reuse = 1;
    if(setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, (const char*)&reuse,
    sizeof(reuse)) < 0) exit(1);
    if(setsockopt(sockfd, SOL_SOCKET, SO_REUSEPORT, (const char*)&reuse,
    sizeof(reuse)) < 0) exit(1);
    if(bind(sockfd, (struct sockaddr*)&local, sizeof(struct sockaddr)) == -1)
        errx(1, "bind(2) failed, socket already bound");
    if(listen(sockfd, BACKLOG) == -1)
        errx(1, "listen(2) failed");
    while(1) {

        if(flood == 1 && (nmex > burst)) continue; /* antiflood */
        if(lazy == 0) {

            unsigned int sin_size = sizeof(struct sockaddr_in);
            if((nsockfd = accept(sockfd, (struct sockaddr *)&remote,
            &sin_size)) == -1) exit(1);
        }
        else {

            close(nsockfd); continue;
        }
        FILE *fr = fopen(TCR, "w");
        if(flood == 1) nmex++;
        if(fr == NULL) exit(1);
        else {

            memset(revbuf, 0, sizeof(revbuf));
            int block = 0;
            while((block = recv(nsockfd, revbuf, RWLENGTH, 0)) > 0) {

                int write = fwrite(revbuf, sizeof(char), block, fr);
                if(write < block) errx(1, "failed write on server");

                memset(revbuf, 0, sizeof(revbuf));
                if(block == 0 || block != RWLENGTH) break;
            }
            if(block < 0) exit(1);
            fseek(fr, 0, SEEK_END);
            long flen = ftell(fr);
            bytein =  bytein + flen;
            fclose(fr);
            int pid = getpid();
            kill(pid, SIGUSR2);
        }
        close(nsockfd);
    }
    return NULL;
}

void keyboard(void) {

    if(nrow  >= 30000) nrow = 0;
    int h = hpos, ccc = -1, visible, pos = nrow;
    for(int i = 0; i <= IN; ++i) {

        if(scrollback == 1) {

            wrefresh(stdscr); redrawwin(txt);
        }
        getyx(mainwin, y, x);
        int cx = getch();
        if(cx == '\n') break;
        scrollback = 0;
        switch(cx) {

        case KEY_PPAGE:

            wrefresh(stdscr);
            getyx(mainwin, y, x);
            getmaxyx(mainwin, row, col);
            visible = row-3;
            if(nrow > 0) {

                scrollback = 1;
                curs_set(0);
                draw();
                if(pos <= 0) {

                    pos = 0;
                    curs_set(1);
                    draw();
                    move(y, x);
                    i = strlen(in);
                    break;
                }
             pos = pos - (visible/2);
             prefresh(pad, pos, 0, 1, 0, visible, (COLS-1));
            }
            curs_set(1);
            draw();
            move(y, x);
            i = strlen(in);
        break;
        case KEY_NPAGE:

            wrefresh(stdscr);
            getyx(mainwin, y, x);
            getmaxyx(mainwin, row, col);
            visible = row-3;
            if(nrow > 0) {

                scrollback = 1;
                curs_set(0);
                draw();
                if(pos >= nrow || pos + visible >= nrow) {

                    pos = nrow - visible;
                    prefresh(pad, pos, 0, 1, 0, visible, (COLS-1));
                    curs_set(1);
                    draw();
                    move(y, x);
                    i = strlen(in);
                    break;
                }
                pos = pos + (visible/2);
                prefresh(pad, pos, 0, 1, 0, visible, (COLS-1));
            }
            curs_set(1);
            draw();
            move(y, x);
            i = strlen(in);
        break;
        case KEY_LEFT:

            getmaxyx(mainwin, row, col);
            if(ccc < 0) {

                move((row-1), 0);
                clrtoeol();
                in[0] = '\0';
                msg[0] = '\0';
                return;
            }
            move((row-1), 0);
            clrtoeol();
            if(ccc != 0) ccc--;
            memcpy(in, commands[ccc], sizeof(in));
            mvprintw((row-1), 0, " =--> %s", commands[ccc]);
            move((row-1), (6+strlen(commands[ccc])));
            getyx(mainwin, y, x);
            draw();
            move(y, x);
            i = strlen(in);
        break;
        case KEY_RIGHT:

            getmaxyx(mainwin, row, col);
            if(ccc > (COMMANDS-1)) {

                move((row-1), 0);
                clrtoeol();
                in[0] = '\0';
                msg[0] = '\0';
                return;
            }
            move((row-1), 0);
            clrtoeol();
            if(ccc != (COMMANDS-1)) ccc++;
            if(ccc == (COMMANDS-1)) ccc--;
            memcpy(in, commands[ccc], sizeof(in));
            mvprintw((row-1), 0, " =--> %s", commands[ccc]);
            move((row-1), (6+strlen(commands[ccc])));
            getyx(mainwin, y, x);
            draw();
            move(y, x);
            i = strlen(in);
        break;
        case KEY_UP:

            getmaxyx(mainwin, row, col);
            if(h < 0) {

                in[0] = '\0';
                msg[0] = '\0';
                return;
            }
            move((row-1), 0);
            clrtoeol();
            if(h != 0) h--;
            memcpy(in, s[h], 61);
            in[IN] = '\0';
            mvprintw((row-1), 0, " =--> %s", in);
            move((row-1), (6+strlen(in)));
            getyx(mainwin, y, x);
            draw();
            move(y, x);
            i = strlen(in);
        break;
        case KEY_DOWN:

            getmaxyx(mainwin, row, col);
            if(h > hpos) {

                in[0] = '\0';
                msg[0] = '\0';
                return;
            }
            move((row-1), 0);
            clrtoeol();
            if(h != hpos) h++;
            memcpy(in, s[h], 61);
            in[IN] = '\0';
            mvprintw((row-1), 0, " =--> %s", in);
            move((row-1), (6+strlen(in)));
            getyx(mainwin, y, x);
            draw();
            move(y, x);
            i = strlen(in);
        break;
        case 127:                                  /* FALLTHROUGH */
        case KEY_BACKSPACE:
        case KEY_CANCEL:
        case '\b':
        case KEY_DC:

            if(x >= 7) {                   /* don't delete prompt */

                getyx(mainwin, y, x);
                move(y, x-1);
                delch();
                in[i] = 0;
                in[i+1] = 0;
                in[i-1] = 0;
            }
            else {

                x = 7;
            }
            if(i >= 0) {         /* overwrite already typed chars */

                i = i - 1;
            }
            if(i < 0) {    /* don't break on backspaces underflow */

                i = 0;
                break;
            }
        break;
        default: break;
    }
    if(!isprint(cx)) i--;
    if(isprint(cx)) {

        if(cx == KEY_PPAGE || cx == KEY_NPAGE || cx == KEY_END ||
           cx == KEY_CANCEL || (i == IN && cx != '\n')) i--;
        else {

            in[i] = cx; in[i+1] = '\0'; addch(cx);
        }
    }
    } /* for loop end here*/
    strlcpy(msg, in, IN+1); scrollback = 0; deleteln(); fflush(stdin);
}

int send_file(void) {

    FILE *f = fopen(msg, "r");
    if(checkfsize(f, msg, MAX_SNDFSIZE, 0) == 1) return(1);
    colorinfo(); wprintw(txt, "sending file %s to %s\n", msg, remo);
    if(grp == 1) {

        char tmp[STR] = {"\0"};
        strlcpy(tmp, msg, sizeof(tmp));
        memset(msg, 0, sizeof(msg));
        char buf[] = " sending a file to group";
        sendinfos(buf);
        sleep(5);
        memset(msg, 0, sizeof(msg));
        strlcpy(msg, tmp, sizeof(msg));
    }
    (grp == 1) ? gcrypt(NULL, NULL, msg) : ucrypt(NULL, NULL, msg);
    return(0);
}

char *deblank(char *str) {

    char *out = str, *put = str;
    for(; *str != '\0'; ++str) {

        if(*str != '>' && *str != '<' && *str != ';' && *str != ' '
        && *str != '`' && *str != '&' && *str != '\\' && *str != '"'
        && *str != '|' && *str != '/' && *str != '\'' && *str != '#'
        && *str != '$' && *str != '*' && *str != '[' && *str != ']'
        && *str != '{' && *str != '}')

            *put++ = *str;
    }
    *put = '\0';
    return out;
}

int showfile(char const *file) {

    FILE *f = NULL;
    if(file == NULL) {

        f = fopen(tcbot, "r");
        if(f == NULL) return(1);
        hm(); c_info(); wprintw(txt, "<*> bot's output:\n"); c_info_off();
    }
    else {

        f = fopen(msg, "r");
        if(f == NULL) return(1);
        unsigned char data[STR] = {"\0"};
        fread(data, 100, 1, f);
        rewind(f);
        if(isbin(data, 100)) {

            colorerror(); wprintw(txt, "%s seems a binary\n", msg);
            hexdump(data);
            fclose(f);
            return(0);
        }
    }
    char *b = NULL;
    size_t len = 0;
    int i = 0;
    while (getline(&b, &len, f) != -1) {

        char *lin = b;
        if(lin == NULL) break;
        int le = strlen(lin);
        lin[COLS - 7] = '\0';         /* 7 is time+1 chars: "12:08 " */
        hm(); wprintw(txt, "%s", lin);
        if(le > (COLS - 7)) wprintw(txt, "\n");
        i++;
        if(i >= LINES - 4) break;
    }
    free(b); fclose(f); return(0);
}

void printkeys(void) {

    hm(); c_info(); wprintw(txt,"<*> pgp keys:\n"); c_info_off();
    for(int i = 0; i < nkey; i++) {

        hm(); wprintw(txt,"  %d.[ ", i);
        ring[i].descr[strcspn(ring[i].descr,"\n")] = '\0';
        c_whit(); wprintw(txt,"%-64s", ring[i].descr); c_whit_off();
        wprintw(txt," ]\n");
    }
    hm(); wprintw(txt,"(total count: %d)\n",nkey);
    loadauthkeys(); printkauth(); printkprv();
}

void xtermresize(int sig) {

    endwin(); refresh(); move(row-1, 0); clrtoeol(); move(row-2, 0); clrtoeol();
    int new_row;
    getmaxyx(mainwin, new_row, col);
    getyx(txt, y, x);
    if(new_row < row && y >= new_row-2) {

        int r = row - new_row;
        /* XXX if r > 3 blank r-3 lines before bottom txt cursor line */
        wscrl(txt, r); refresh();
    }
    row = new_row;
    wresize(mainwin, row, col);
    wresize(txt, row-2, col);
    refresh(); draw(); redrawwin(txt); refresh();
}

void torcircuit(void) {

    waitingfor();
    hm(); c_info(); wprintw(txt, "<*> circuit:\n"); c_info_off();
    alive();
    hm(); wprintw(txt, "real: ");
    c_whit(); wprintw(txt, "%s", chk); c_whit_off();
    wprintw(txt, " exit node: ");
    c_whit(); wprintw(txt, "%s\n", chktor); c_whit_off();
    hm(); wprintw(txt, "host: %s\n", hidden());
}

void createtcdirs(void) {

    FILE *d = NULL;
    d = fopen(tcdir, "r"); if(d == NULL)
    errx(1,"tc's dir not found: \"mkdir %s\" and restart", tcdir);
    else fclose(d); fflush(d);

    d = fopen(tctordir, "r"); if(d == NULL) {
        if(mkdir((const char *)tctordir, 0700) != 0)
        fprintf(stderr, "error: %d making %s\n", errno, tctordir);
    } else fclose(d); fflush(d);

    d = fopen(tcauthdir, "r"); if(d == NULL) {
        if(mkdir((const char *)tcauthdir, 0700) != 0)
        fprintf(stderr, "error: %d making %s\n", errno, tcauthdir);
    } else fclose(d); fflush(d);

    d = fopen(tchiddendir, "r"); if(d == NULL) {
        if(mkdir((const char *)tchiddendir, 0700) != 0)
        fprintf(stderr, "error: %d making %s\n", errno, tchiddendir);
    } else fclose(d); fflush(d);

    d = fopen(tcauthclidir, "r"); if(d == NULL) {
        if(mkdir((const char *)tcauthclidir, 0700) != 0)
        fprintf(stderr, "error: %d making %s\n", errno, tcauthclidir);
    } else fclose(d); fflush(d);
}

void createrc(void) {

    FILE *f = fopen(tcrc, "w");
    char buf[MSG] = {"\0"};
    strlcat(buf, home, sizeof(buf));
    strlcat(buf, "/.tc", sizeof(buf));
    fprintf(f, "Log notice syslog\n");
    fprintf(f, "RunAsDaemon 1\n");
    fprintf(f, "DataDirectory %s/tor\n", buf);
    fprintf(f, "HiddenServiceDir %s/tor/hidden_service\n", buf);
    fprintf(f, "HiddenServicePort 2323 127.0.0.1:2323\n");
    fprintf(f, "ClientOnionAuthDir %s/tor/auth\n", buf);
    fclose(f);
}

void createtcconf(void) {

    FILE *f = fopen(tcconf, "w");
    fprintf(f, "# passphrase\n");
    fprintf(f, "''\n");
    fprintf(f, "# Tor proxy address and port\n");
    fprintf(f, "127.0.0.1:9050\n");
    fprintf(f, "# tc address and port\n");
    fprintf(f, "127.0.0.1:2323\n");
    fprintf(f, "# colors\n");
    fprintf(f, "2:0\n");
    fprintf(f, "0:7\n");
    fprintf(f, "7:0\n");
    fprintf(f, "6:0\n");
    fprintf(f, "# contacts\n");
    fprintf(f, "tc:84BE7F4663CE2B5A:%s\n",TCURL);
    fprintf(f, "# groups\n");
    fprintf(f, "Ramones:Joey,DeeDee,Johnny,Tommy,\n");
    fprintf(f, "JAZZ:Blakey,Gillespie,Telonious,Davis,Parker,\n");
    fprintf(f, "# eof\n");
    fclose(f);
}

void invalidnick(char *st) {

    char bad[] = "_badnick_";
    if(strstr(st, "à") != NULL) strlcpy(nick, bad, sizeof(nick));
    if(strstr(st, "è") != NULL) strlcpy(nick, bad, sizeof(nick));
    if(strstr(st, "ì") != NULL) strlcpy(nick, bad, sizeof(nick));
    if(strstr(st, "ò") != NULL) strlcpy(nick, bad, sizeof(nick));
    if(strstr(st, "ù") != NULL) strlcpy(nick, bad, sizeof(nick));
    if(strstr(st, "<*>") != NULL) strlcpy(nick, bad, sizeof(nick));
    if(strstr(st, "<*") != NULL) strlcpy(nick, bad, sizeof(nick));
    if(strstr(st, "*>") != NULL) strlcpy(nick, bad, sizeof(nick));
    if(strstr(st, "<decoy>") != NULL) strlcpy(nick, bad, sizeof(nick));
    if(strcmp(st, "decoy") == 0) strlcpy(nick, bad, sizeof(nick));
    if(strcmp(st, "*") == 0) strlcpy(nick, bad, sizeof(nick));
}

void upgrade(void) {

    fprintf(stderr, "new release avaliable, install now? [y/n]: ");
    int ch = getchar();
    if(ch == 'y') {

        FILE *f = NULL;
        f = fopen(tctgz, "r");
        if(f != NULL) { fclose(f); unlink(tctgz); }
        char d[STR] = {"\0"}; /* download */
        strlcat(d, socks4a, sizeof(d));
        strlcat(d, TCURL, sizeof(d));
        strlcat(d, "/tc.tgz -o ", sizeof(d));
        strlcat(d, tctgz, sizeof(d));
        system(d); sleep(3);
        f = NULL;
        f = fopen(tctgz, "r");
        if(f != NULL) {

            fclose(f);
            char i[STR] = {"\0"}; /* install */
            strlcat(i, "tar xzvf ", sizeof(i));
            strlcat(i, tctgz, sizeof(i));
            strlcat(i, " -C ", sizeof(i));
            strlcat(i, tcdir, sizeof(i));
            strlcat(i, "/; ", sizeof(i));
            strlcat(i, tcdir, sizeof(i));
            strlcat(i, "/./tc.sh install", sizeof(i));
            system(i);
        }
        else
        fprintf(stderr, "can't find %s\n", tctgz);
    }
    exit(0);
}

void update(void) {

    #ifdef openbsd
    if(pledge("stdio rpath wpath cpath proc exec", NULL) == -1)
        err(1, "pledge");
    #endif

    torpid();
    if(tor == 1) {

        FILE *f = fopen(tcver, "r");
        if(f != NULL) {

            char localver[65] = {"\0"};
            fgets(localver, 65, f);
            fclose(f);
            char buf[STR] = {"\0"};
            strlcat(buf, socks4a, sizeof(buf));
            strlcat(buf, TCURL, sizeof(buf));
            strlcat(buf, "/tc.ver", sizeof(buf));
            FILE *p = NULL;
            if((p = popen(buf, "r")) != NULL) {

                char rolling[65] = {"\0"};
                fgets(rolling, 65, p); rolling[strcspn(rolling, "\n")] = '\0';
                pclose(p);
                if(strcmp(localver, rolling) == 0) errx(1, "is up to date!");
                else upgrade();
            }
        }
        else upgrade();
    }
    else errx(1, "Tor is down, can't check update");
}

char *mkrandom(void) {

    char *otp = (char *) calloc(1, 16);
    if(otp == NULL) { free(otp); return NULL; }
    FILE *p = NULL;
    if((p = popen("gpg --armor --gen-random 0 16", "r")) != NULL) {

        fgets(otp, N+6, p);
        otp[strcspn(otp, "\n")] = '\0';
        pclose(p);
        deblank(otp);
        return otp;
    }
    else { free(otp); return NULL; }
}

void setsympass(void) {

m00t:
    mvprintw((row-1),0," =--> [*] symmetric encrypt with (g)pg or (o)penssl: ");
    move(row-1, 53);
    echo();
    getnstr(msg, 1);
    noecho();
    curs_set(0);
    deleteln();
    switch((int)msg[0]) {
    case 'g': symg = 1; break;
    case 'o': symo = 1; break;
    default: goto m00t;
    }
w00t:
    curs_set(1);
    mvprintw((row-1),0," =--> [*] (s)et password or use (o)tp: ");
    move(row-1, 39);
    echo();
    getnstr(msg, 1);
    noecho();
    curs_set(0);
    char *otp = NULL;
    switch((int)msg[0]) {
    case 's':
        deleteln();
        curs_set(1);
        mvprintw((row-1),0, " =--> [*] set passphrase: ");
        move(row-1, 26);
        echo();
        getnstr(msg, N+5);
        noecho();
        curs_set(0);
        otp = deblank(msg);
        memset(K, 0, sizeof(K));
        strlcpy(K, otp, sizeof(K));
        deleteln();
        if(msg[0] == '\0') goto w00t;
        break;
    case 'o':
        deleteln();
        memset(K, 0, sizeof(K));
        otp = mkrandom();
        strlcpy(K, otp, sizeof(K));
        break;
    default: goto w00t;
    }
    if(symg == 1) {

        colorinfo();
        wprintw(txt, "symmetric encryption: ON (gpg key: %s)\n", otp);
    }
    if(symo == 1) {

        colorinfo();
        wprintw(txt, "symmetric encryption: ON (openssl key: %s)\n", otp);
    }
    draw();
    if(con == 1 && noblank == 0) {

        char buf[STR] = " activated symmetric ";
        if(symg == 1)
        strlcat(buf, "GPG", sizeof(buf));
        if(symo == 1)
        strlcat(buf, "OPENSSL", sizeof(buf));
        strlcat(buf, " with key: ", sizeof(buf));
        strlcat(buf, otp, sizeof(buf));
        sendinfos(buf);
    }
}

void setcolorz(void) {

    hm(); c_info(); wprintw(txt, "<*> colors:"); c_info_off();
    for(int i = 0; i < (9-1); i++) {

        if(i%4 == 0) {

            wprintw(txt,"\n"); hm();
        }
        wprintw(txt,"  %d.[ ", i);
        wprintw(txt, "%-7s", colors[i]); c_whit_off();
        wprintw(txt," ]");
    }
    wprintw(txt, "\n"); hm(); wprintw(txt, "(total count: %d)\n", (9-1));
    draw();

    int i = -1;
    mvprintw((row-1),0, " =--> [*] set info color: ");
    move(row-1, 27); echo(); getnstr(msg, 1); noecho();
    curs_set(0); deleteln();
    if(msg[0] != '\0' && isdigit(msg[0])) i = atoi(msg);
    if(i >= 0 && i <= 7) {

        info_f = i;
        init_pair(1,info_f,info_b);
        wattron(txt, A_BOLD | COLOR_PAIR(1));
        c_info_off();
    }
    i = -1;
    mvprintw((row-1),0, " =--> [*] set warning color: ");
    move(row-1, 30); echo(); getnstr(msg, 1); noecho();
    curs_set(0); deleteln();
    if(msg[0] != '\0' && isdigit(msg[0])) i = atoi(msg);
    if(i >= 0 && i <= 7) {

        erro_f = i;
        init_pair(4,erro_f,erro_b);
        wattron(txt, A_BOLD | COLOR_PAIR(4));
        c_erro_off();
    }
    i = -1;
    mvprintw((row-1),0, " =--> [*] set text color: ");
    move(row-1, 27); echo(); getnstr(msg, 1); noecho();
    curs_set(0); deleteln();
    if(msg[0] != '\0' && isdigit(msg[0])) i = atoi(msg);
    if(i >= 0 && i <= 7) {

        text_f = i;
        init_pair(3,text_f,text_b);
        wattron(txt, A_NORMAL | COLOR_PAIR(3));
        c_text_off();
    }
    i = -1;
    mvprintw((row-1),0, " =--> [*] set bars foregroud color: ");
    move(row-1, 37); echo(); getnstr(msg, 1); noecho();
    curs_set(0); deleteln();
    if(msg[0] != '\0' && isdigit(msg[0])) i = atoi(msg);
    if(i >= 0 && i <= 7) {

        bars_f = i;
        init_pair(2,bars_f,bars_b); color_set(2,NULL);
        c_prompt();
    }
    i = -1;
    mvprintw((row-1),0, " =--> [*] set bars backgroud color: ");
    move(row-1, 37); echo(); getnstr(msg, 1); noecho();
    curs_set(0); deleteln();
    if(msg[0] != '\0' && isdigit(msg[0])) i = atoi(msg);
    if(i >= 0 && i <= 7) {

        bars_b = i;
        init_pair(2,bars_f,bars_b); color_set(2,NULL);
        c_prompt();
    }
}

void chat(void) {

    if((mainwin = initscr()) == NULL) exit(1);
    getmaxyx(mainwin, row, col);
    noecho();
    raw();
    txt = newwin(row-2, col, 1, 0);
    scrollok(txt, TRUE);
    pad = newpad((row+200), (COLS-1));
    keypad(pad, TRUE);
    clear();
    mutual(0);
    start_color();
    draw();
    keypad(mainwin, TRUE);
    struct sigaction act1, act2, act3;
    act1.sa_handler = &decrypt; act1.sa_flags = SA_RESTART;
    act2.sa_handler = &cleanup; act2.sa_flags = SA_RESTART;
    act3.sa_handler = &xtermresize; act3.sa_flags = SA_RESTART;
    sigaction(SIGUSR2, &act1, NULL);
    sigaction(SIGINT, &act2, NULL);
    sigaction(SIGWINCH, &act3, NULL);
    unlink(TCS); shred(TCT); unlink(TCR); shred(TCN);
    pthread_create(&clk, NULL, clocker, NULL);
    pthread_detach(pthread_self());
    pthread_create(&srv, NULL, listener, NULL);
    pthread_detach(pthread_self());
    invalidnick(nick);
    torpid();
    tstart = time(&t);
    strlcpy(cipher, "AES256", sizeof(cipher));
    while (1) {

        deleteln();
        draw();
        redrawwin(txt);
        curs_set(1);
        keyboard();
        if(msg[0] == '\0') continue;
        history(msg);
        in[0] = '\0';
        if(strcmp(msg, "/algo") == 0) {

            algoz(); continue;
        }
        if((strcmp(msg, "/away") == 0) || (strcmp(msg, "/a") == 0)) {

            if(away == 0) {

                mvprintw((row-1),0, " =--> [*] set away message: ");
                move(row-1, 28);
                echo();
                getnstr(way, N*4);
                noecho();
                curs_set(0);
                if(way[0] == '\0') {

                    strlcpy(way, " ", 1);
                }
                colorinfo();
                wprintw(txt, "you have been marked as being away\n");
                away = 1; logs = 1; continue;
            }
            if(away == 1) {

                colorinfo();
                wprintw(txt, "you are no longer marked as being away\n");
                away = 0; logs = 0; continue;
            }
        }
        if((strcmp(msg, "/bot") == 0) || (strcmp(msg, "/b") == 0)) {

            if(con == 0) {

                nocom(); continue;
            }
            if(con == 1) {

                waitingfor();
                FILE *b = NULL;
                b = fopen(tcbotsh, "r");
                if(b != NULL) {

                    fclose(b);
                    char buf[STR] = {"\0"};
                    strlcat(buf, tcbotsh, sizeof(buf));
                    strlcat(buf, " >/dev/null 2>&1", sizeof(buf));
                    system(buf);
                    if(showfile(NULL) == 1) {

                        colorerror(); wprintw(txt, ".bot output not found\n");
                    }
                }
                else {

                    colorerror(); wprintw(txt, "bot.sh file not found\n");
                }
                continue;
            }
        }
        if(strcmp(msg, "/call") == 0) {

            if(con == 1) {

                colorerror(); wprintw(txt, "already joined to %s\n", remo);
                continue;
            }
            mvprintw((row-1),0, " =--> [*] onion: ");
            move(row-1, 17);
            echo();
            getnstr(msg, 63-1);
            noecho();
            if(msg[0] == '\0' || strlen(msg) != 62 || msg[56] != '.' ||
               msg[57] != 'o' || msg[58] != 'n' || msg[59] != 'i' ||
               msg[60] != 'o' || msg[61] != 'n') {

                   colorerror(); wprintw(txt, "invalid hostname\n");
                   continue;
            }
            msg[strcspn(msg, "\n")] = '\0';
            deleteln();
            strlcpy(callo, msg, sizeof(callo));
            o[0] = callo;
            msg[0] = '\0';
            if(sym == 1) setsympass();
            mvprintw((row-1),0, " =--> [*] pgp keyid: ");
            move(row-1, 21);
            echo();
            getnstr(msg, 17-1);
            noecho();
            if(msg[0] == '\0' || strlen(msg) != 16 ||
               checkkey("",msg) != 0) {

                   colorerror(); wprintw(txt, "invalid pgp key\n");
                   continue;
            }
            msg[strcspn(msg, "\n")] = '\0';
            deleteln();
            strlcpy(remo, callo, sizeof(remo));
            strlcpy(callk, msg, sizeof(callk));
            k[0] = callk;
            colorinfo(); wprintw(txt, "joining %s", o[0]);
            waitingfor(); who++; con = 1;
            char buf[] = " has joined the conversation";
            sendinfos(buf); continue;
        }
        if(strcmp(msg, "/camuflage") == 0) {

            if(con == 0) {

                nocom(); continue;
            }
            if(camuflage == 0) {

                colorinfo(); wprintw(txt, "camuflage: ON\n");
                camuflage = 1; continue;
            }
            if(camuflage == 1) {

                colorinfo(); wprintw(txt, "camuflage: OFF\n");
                camuflage = 0; continue;
            }
        }
        if(strcmp(msg, "/circuit") == 0 || (strcmp(msg, "/c") == 0)) {

            torcircuit(); torpid(); continue;
        }
        if(strcmp(msg, "/clear") == 0) {

            clear();
            delwin(txt);
            getmaxyx(mainwin, row, col);
            txt = newwin(row-2, col, 1, 0);
            scrollok(txt, TRUE);
            continue;
        }
        if(strcmp(msg, "/colors") == 0) {

            setcolorz(); continue;
        }
        if(strcmp(msg, "/date") == 0) {

            time(&t);
            tm = *localtime(&t);
            colorinfo(); wprintw(txt, "%s", ctime(&t));
            continue;
        }
        if(strcmp(msg, "/decoy") == 0 || (strcmp(msg, "/d") == 0)) {

            if(con == 0) {

                nocom(); continue;
            }
            if(con == 1 && decoy == 0) {

                pthread_create(&dec, NULL, send_decoy, NULL);
                pthread_detach(pthread_self());
                colorinfo(); wprintw(txt, "decoy traffic: ON\n");
                decoy = 1; continue;
            }
            if(con == 1 && decoy == 1) {

                pthread_cancel(dec);
                colorinfo(); wprintw(txt, "decoy traffic: OFF\n");
                decoy = 0; continue;
            }
        }
        if(strcmp(msg, "/disconnect") == 0) {

            colorinfo();
            wprintw(txt, "disconnecting Tor, killing process\n");
            waitingfor();
            if(killtor() == 1) continue;
            nocom(); mutual(0); continue;
        }
        if((strcmp(msg, "/exit") == 0) || (strcmp(msg, "/e") == 0)) {

            cleanup(0);
        }
        if((strcmp(msg, "/flags") == 0)) {

            flagz(); continue;
        }
        if((strcmp(msg, "/flood") == 0) || (strcmp(msg, "/f") == 0)) {

            if(flood == 0) {

                mvprintw((row-1),0,
                " =--> [*] set message rate per minute (10-50): ");
                move(row-1, 47);
                echo();
                getnstr(msg, 2);
                noecho();
                if(msg[0] == '\0') continue;
                if(isdigit(msg[0]) == 0 || isdigit(msg[1]) == 0) continue;
                burst = (atoi(msg));
                if(burst < 10 || burst > 50) {

                    burst = 0; continue;
                }
                colorinfo();
                wprintw(txt,
                "antiflood: ON, print max %d message per minute\n", burst);
                flood = 1;
                continue;
            }
            if(flood == 1) {

                colorinfo(); wprintw(txt, "antiflood: OFF\n");
                flood = 0; burst = 0; continue;
            }
        }
        if(strcmp(msg, "/generate") == 0) {

            mvprintw((row-1),0, " =--> [*] generate new .onion? [y/n]: ");
            move(row-1, 38);
            echo();
            getnstr(msg, 1);
            noecho();
            if(msg[0] != 'y') continue;
            else {

                waitingfor();
                colorinfo(); wprintw(txt, "generating new .onion\n");
                if(saveonion() == 1) continue;
                fireuptor(); hm(); wprintw(txt, "new: %s\n", hidden());
                nocom(); continue;
            }
        }
        if((strcmp(msg, "/groups") == 0) || (strcmp(msg, "/g") == 0)) {

            showgroups(); continue;
        }
        if(strcmp(msg, "/help") == 0 || (strcmp(msg, "/h") == 0)) {

            help(); continue;
        }
        if(strcmp(msg, "/history") == 0) {

            history(NULL); continue;
        }
        if((strcmp(msg, "/info") == 0) || (strcmp(msg, "/i") == 0)) {

            hm(); c_info(); wprintw(txt, "<*> info:\n"); c_info_off();
            hm(); wprintw(txt, "real address: %15s", chk);
            wprintw(txt, "\t\ttraffic in: %17ld\n", bytein);
            hm(); wprintw(txt, "tc daemon: %11s :%5d", TCDAEMON, TCPORT);
            wprintw(txt, "\t\ttraffic out: %16ld\n", byteout);
            hm(); wprintw(txt, "Tor proxy: %11s :%5d", TPROXY, TPORT);
            wprintw(txt, "\t\tantiflood level: %12d\n", burst);
            hm(); wprintw(txt, "Tor address: %16s", chktor);
            wprintw(txt, "\t\tlast minute mex: %12d\n", nmex);
            hm(); wprintw(txt, "Tor version: %16s", torver);
            wprintw(txt, "\t\tremote peers: %15d\n", who);
            hm(); wprintw(txt, "Tor process: %16d", tpid);
            wprintw(txt, "\t\tdecoys: %21d\n", ndec);
            hm(); wprintw(txt, "Tor keepalive minute: %7d", ALIVE-interval);
            wprintw(txt, "\t\tscrollback lines: %11d\n",nrow);
            hm(); wprintw(txt, "gnupg version: %14s", gpgver);
            wprintw(txt, "\t\tsymmetric gpg: %14d\n", symg);
            hm(); wprintw(txt, "lines x cols: %9d x%4d", LINES, COLS);
            wprintw(txt, "\t\tsymmetric openssl: %10d\n", symo);
            continue; /* notice: data is padded to 29 chars */
        }
        if(strcmp(msg, "/join") == 0 || (strcmp(msg, "/j") == 0)) {

            if(con == 0 && sym == 1) setsympass();
            if(con == 1) {

                colorerror(); wprintw(txt, "already joined to %s\n", remo);
                continue;
            }
            mvprintw((row-1),0, " =--> [*] join to: ");
            move(row-1, 19);
            echo();
            getnstr(msg, N-1);
            noecho();
            if(msg[0] == '\0') continue;
            deleteln();
            if(checkname(msg) == 1) continue;
            waitingfor();
            colorinfo(); wprintw(txt, "joining %s\n", msg);
            for(int i = 0; i < who; i++) {

                hm(); wprintw(txt, "%d. %s\n", i+1, o[i]);
                wrefresh(mainwin);
            }
            con = 1;
            char buf[] = " has joined the conversation";
            sendinfos(buf); continue;
        }
        if((strcmp(msg, "/keys") == 0) || (strcmp(msg, "/k") == 0)) {

            printkeys(); continue;
        }
        if(strcmp(msg, "/lazy") == 0) {

            if(lazy == 0) {

                colorinfo(); wprintw(txt, "laziness: ON\n");
                lazy = 1; continue;
            }
            if(lazy == 1) {

                colorinfo(); wprintw(txt, "laziness: OFF\n");
                lazy = 0; continue;
            }
        }
        if((strcmp(msg, "/log") == 0) || (strcmp(msg, "/l") == 0)) {

            if(logs == 0) {

                colorinfo(); wprintw(txt, "logging: ON\n");
                logs = 1; continue;
            }
            if(logs == 1) {

                colorinfo(); wprintw(txt, "logging: OFF\n");
                logs = 0; continue;
            }
        }
        if(strcmp(msg, "/me") == 0 || (strcmp(msg, "/m") == 0)) {

            if(con == 0) {

                nocom(); continue;
            }
            else {

                mvprintw((row-1),0, " =--> [*] notice: ");
                move(row-1, 18);
                echo();
                getnstr(msg, N*4);
                noecho();
                curs_set(0);
                if(msg[0] == '\0') continue;
                deleteln();
                wprintw(txt, "%02d:%02d ", tm.tm_hour, tm.tm_min);
                c_whit(); wprintw(txt, "*notice* "); c_whit_off();
                c_erro(); wprintw(txt, "%s %s\n", nick, msg);
                c_erro_off();
                char buf[STR] = {"\0"};
                strlcat(buf, "<notice> ", sizeof(buf));
                strlcat(buf, nick, sizeof(buf));
                strlcat(buf, " ", sizeof(buf));
                strlcat(buf, msg, sizeof(buf));
                strlcpy(msg, buf, sizeof(msg));
                forward();
                addscrollback(msg);
                continue;
            }
        }
        if(strcmp(msg, "/names") == 0) {

            showcontacts(NULL); continue;
        }
        if((strcmp(msg, "/nick") == 0) || (strcmp(msg, "/n") == 0)) {

            mvprintw((row-1),0, " =--> [*] new nick: ");
            move(row-1, 20);
            echo();
            getnstr(msg, N-1);
            noecho();
            curs_set(0);
            if(msg[0] == '\0') continue;
            deleteln();
            if(strcmp(msg, nick) == 0) continue;
            char tmp[STR] = {"\0"};
            strlcpy(tmp, nick, sizeof(tmp));
            memset(nick, 0, sizeof(nick));
            strlcpy(nick, msg, sizeof(nick));
            invalidnick(nick);
            if(con == 1) {

                char buf[STR] = {"\0"};
                strlcat(buf, tmp, sizeof(buf));
                strlcat(buf, " has changed nick to ", sizeof(buf));
                strlcat(buf, nick, sizeof(buf));
                strlcpy(msg, buf, sizeof(msg));
                forward();
            }
            continue;
        }
        if(strcmp(msg, "/part") == 0 || (strcmp(msg, "/p") == 0)) {

            if(con == 0) {

                nocom(); continue;
            }
            else {

                mvprintw((row-1),0, " =--> [*] set leave message: ");
                move(row-1, 29);
                echo();
                getnstr(way, N*3);
                noecho();
                curs_set(0);
                if(way[0] == '\0') {

                    char buf[] = " has left the conversation";
                    sendinfos(buf);
                    nocom(); mutual(0); continue;
                }
                else {

                    char buf[STR] = " has left the conversation: ";
                    strlcat(buf, way, sizeof(buf));
                    sendinfos(buf);
                    nocom(); mutual(0); continue;
                }
            }
        }
        if((strcmp(msg, "/quit") == 0) || (strcmp(msg, "/q") == 0)) {

            if(con == 0) cleanup(0);
            mvprintw((row-1),0, " =--> [*] set quit message: ");
            move(row-1, 28);
            echo();
            getnstr(way, N*3);
            curs_set(0);
            if(way[0] == '\0') {

                char buf[] = " has quit the conversation";
                sendinfos(buf);
                pthread_cancel(srv);
                if(decoy == 1) pthread_cancel(dec);
                cleanup(0);
            }
            else {

                char buf[STR] = " has quit the conversation: ";
                strlcat(buf, way, sizeof(buf));
                sendinfos(buf);
                pthread_cancel(srv);
                if(decoy == 1) pthread_cancel(dec);
                cleanup(0);
            }
        }
        if((strcmp(msg, "/restore") == 0) || (strcmp(msg, "/r") == 0)) {

            hm(); c_info(); wprintw(txt, "<*> restore:\n"); c_info_off();
            char buf[STR] = {"\0"};
            strlcat(buf, "ls ", sizeof(buf));
            strlcat(buf, home, sizeof(buf));
            strlcat(buf, "/.tc/ | grep onion | cut -c 5-66", sizeof(buf));
            FILE *p = popen(buf, "r");
            if(p == NULL) pclose(p);
            else {

                int rest = 0;
                while(p != NULL) {

                    fgets(restore[rest].onion, 64, p);
                    if(restore[rest].onion[0] == '\0') break;
                    restore[rest].onion[strcspn(restore[rest].onion,"\n")]='\0';
                    rest++;
                    if(rest >= N) break;
                }
                pclose(p);
                for(int i = 0; i < rest; i++) {

                    hm(); wprintw(txt, "  %d.[  ", i); c_whit();
                    wprintw(txt, "%-63s", restore[i].onion); c_whit_off();
                    wprintw(txt, " ]\n");
                }
                hm(); wprintw(txt, "(total count: %d)\n", rest);
                draw();
                if(rest == 0) continue;
                mvprintw((row-1),0, " =--> [*] choose .onion to restore: ");
                move(row-1, 36);
                echo();
                getnstr(msg, 1);
                noecho();
                if(msg[0] == '\0') continue;
                if(!isdigit(msg[0])) continue;
                int on = (atoi(msg));
                if(on >= 0 && on < rest) {

                    waitingfor();
                    colorinfo(); wprintw(txt, "restoring .onion\n");
                    if(saveonion() == 1) continue;
                    buf[0] = '\0';
                    strlcat(buf, "mv ", sizeof(buf));
                    strlcat(buf, home, sizeof(buf));
                    strlcat(buf, "/.tc/tor.", sizeof(buf));
                    strlcat(buf, restore[on].onion, sizeof(buf));
                    strlcat(buf, " ", sizeof(buf));
                    strlcat(buf, home, sizeof(buf));
                    strlcat(buf, "/.tc/tor", sizeof(buf));
                    system(buf);
                    fireuptor();
                    hm(); wprintw(txt, "new: %s\n", hidden());
                    nocom();
                }
            }
            continue;
        }
        if(strcmp(msg, "/send") == 0) {

            if(con == 0) {

                nocom(); continue;
            }
            else {

                mvprintw((row-1),0, " =--> [*] file: ");
                move(row-1, 16);
                echo();
                getnstr(msg, MSG);
                noecho();
                curs_set(0);
                if(msg[0] == '\0') continue;
                if(checkfile() == 1) continue;
                if(send_file() == 1) {
                    colorerror();
                    wprintw(txt, "can't send!\n");
                }
                continue;
            }
        }
        if(strcmp(msg, "/show") == 0 || (strcmp(msg, "/s") == 0)) {

            mvprintw((row-1),0, " =--> [*] file: ");
            move(row-1, 16);
            echo();
            getnstr(msg, MSG);
            noecho();
            curs_set(0);
            if(msg[0] == '\0') continue;
            if(checkfile() == 1) continue;
            hm(); c_info(); wprintw(txt, "<*> %s:\n", msg); c_info_off();
            if(showfile(msg) == 1) continue;
            continue;
        }
        if(strcmp(msg, "/strict") == 0) {

            mvprintw((row-1),0,
            " =--> [*] authorization (c)lient (s)erver (m)utual: ");
            move(row-1, 52);
            echo();
            getnstr(msg, 1);
            noecho();
            if(msg[0] == 'c' || msg[0] == 's' || msg[0] == 'm') {

                waitingfor(); strictness(); loadauthkeys();
                if(nauth >= 1 || nprv >= 1) {

                    move(row-1, 0); clrtoeol();
                    mvprintw((row-1),0, " =--> [*] restart Tor? [y/n]: ");
                    move(row-1, 30);
                    echo();
                    getnstr(msg, 1);
                    noecho();
                    if(msg[0] != 'y') continue;
                    else {

                        if(killtor() == 1) continue;
                        if(con == 0) nocom();
                        fireuptor();
                        if(strict == 1) {

                            colorinfo();
                            wprintw(txt,"onion authorization: ON\n");
                        }
                        else {

                            colorinfo();
                            wprintw(txt,"onion authorization: OFF\n");
                        }
                    continue;
                    }
                }
            }
            continue;
        }
        if(strcmp(msg, "/symmetric") == 0) {

            if(con == 0 && sym == 0) {

                nocom(); continue;
            }
            if(con == 0 && sym == 1) {

                colorinfo(); wprintw(txt, "symmetric encryption: OFF\n");
                sym = 0;  symg = 0; symo = 0; nocom(); continue;
            }
            if(con == 1 && sym == 0) {

                setsympass(); sym = 1; continue;
            }
            if(con == 1 && sym == 1) {

                char buf[] = " has switched OFF symmetric encyption";
                sendinfos(buf);
                colorinfo(); wprintw(txt, "symmetric encryption: OFF\n");
                sym = 0; symg = 0; symo = 0; continue;
            }
        }
        if((strcmp(msg, "/topic") == 0) || (strcmp(msg, "/t") == 0)) {

            mvprintw((row-1),0, " =--> [*] new topic: ");
            move(row-1, 21);
            echo();
            getnstr(msg, N*3);
            noecho();
            curs_set(0);
            if(msg[0] == '\0') continue;
            topic(msg);
            if(con == 1) {

                char buf[STR] = {"\0"};
                strlcat(buf, nick, sizeof(buf));
                strlcat(buf, " changed the topic to: ", sizeof(buf));
                strlcat(buf, msg, sizeof(buf));
                strlcpy(msg, buf, sizeof(msg));
                forward();
            }
            continue;
        }
        if((strcmp(msg, "/uptime") == 0) || (strcmp(msg, "/u") == 0)) {
            /* ripped from top(1)'s display.c */
            time_t uptime;
            unsigned int d, h, m, se;
            tstop = time(&t);
            uptime = difftime(tstop,tstart);
            d = uptime / (3600 * 24);
            uptime %= (3600 * 24);
            h = uptime / 3600;
            uptime %= 3600;
            m = uptime / 60;
            se = uptime % 60;
            colorinfo();
            if(d > 1 || d == 0)
            wprintw(txt, "tc uptime: %u days, %02u:%02u:%02u\n", d, h, m, se);
            if(d == 1)
            wprintw(txt, "tc uptime: %u day, %02u:%02u:%02u\n", d, h, m, se);
            continue;
        }
        if((strcmp(msg, "/voice") == 0) || (strcmp(msg, "/v") == 0)) {

            if(con == 0) {

                nocom(); continue;
            }
            if(con == 1 && voice == 0) {

                colorinfo(); wprintw(txt, "text to speech: ON\n"); voice = 1;
                continue;
            }
            if(con == 1 && voice == 1) {

                colorinfo(); wprintw(txt, "text to speech: OFF\n"); voice = 0;
                continue;
            }
        }
        if((strcmp(msg, "/who") == 0) || (strcmp(msg, "/w") == 0)) {

            if(con == 1) {

                hm();
                c_info(); wprintw(txt, "<*> talking to:\n"); c_info_off();

                for(int i = 0; i < who; i++) {

                    hm(); wprintw(txt, "%d. %s\n", i+1, o[i]);
                    wrefresh(mainwin);
                }
            }
            if(con == 0) nocom();
            continue;
        }
        if(strcmp(msg, "/whois") == 0) {

            mvprintw((row-1),0, " =--> [*] contact: ");
            move(row-1, 19);
            echo();
            getnstr(msg, N-1);
            noecho();
            curs_set(0);
            if(msg[0] == '\0') continue;
            showcontacts(msg);
            continue;
        }
        if(con == 0) continue;
        redrawwin(txt);
        c_text();
        hm(); wprintw(txt,"<%s> %s\n", nick, msg);
        c_text_off();
        draw();
        addscrollback(NULL);
        if(logs == 1) logger(nick, msg);
        (grp == 1) ? gcrypt(nick, msg, NULL) : ucrypt(nick, msg, NULL);
    }
}

void localfilez(void) {

    if(strlen(home) > MSG) errx(1, "$HOME is too long");
    if(strlen(trm) > MSG) errx(1, "$TERM is too long");
    char v[175][175]; /* 175*175=30625 write jibberish if passed */
    char *path[NFILE] = {"/.tc/.tcs","/.tc/.tct","/.tc/.tcr","/.tc/.tcn",
    "/.tc/tc.conf","/.tc/tc.log","/.tc/.lock","/.tc/.bot","/.tc","/.tc/tc.rc",
    "/.tc/tor","/.tc/tor/hidden_service/hostname","/.tc/tor/auth",
    "/.tc/tor/hidden_service/authorized_clients","/.tc/tc_XXXXXXXXXXXXXX",
    "/.tc/tc.ver","/.tc/tor/hidden_service","/.tc/tc.tgz","/.tc/bot.sh",};
    unsigned long l = 0;
    for(int i = 0; i < NFILE; i++) {

        char buf[STR] = {"\0"};
        strlcat(buf, home, sizeof(buf));
        strlcat(buf, path[i], sizeof(buf));
        strlcpy(v[i], buf, sizeof(v[i]));
        l = l + sizeof(buf);
        assert(l < (sizeof(v) - 1));
        if(l > (sizeof(v) - 1000)) errx(1, "path overflow detected: %lu",l);
    }
    TCS = v[0]; TCT = v[1]; TCR = v[2]; TCN = v[3]; tcconf = v[4]; tclog = v[5];
    tclock = v[6]; tcbot = v[7]; tcdir = v[8]; tcrc = v[9]; tctordir = v[10];
    tchost = v[11]; tcauthdir = v[12]; tcauthclidir = v[13]; tcrcvfile = v[14];
    tcver = v[15]; tchiddendir = v[16]; tctgz = v[17]; tcbotsh = v[18];
}

int main(int argc, char *argv[]) {

    fprintf(stderr, "%s\n", tag);
    if(argc < 2) usage();
    if((home = getenv("HOME")) == 0) exit(1);
    if((trm = getenv("TERM")) == 0) exit(1);

    #ifdef openbsd
    unveil("/dev/null", "w");
    unveil("/bin", "r");                   unveil("/usr/share/terminfo", "r");
    unveil("/bin/ls", "x");                unveil("/bin/mv", "x");
    unveil("/bin/ps", "x");                unveil("/bin/sh", "x");
    unveil("/bin/tar", "x");               unveil("/usr/bin/cut", "x");
    unveil("/usr/bin/awk", "x");           unveil("/usr/bin/grep", "x");
    unveil("/usr/bin/head", "x");          unveil("/usr/bin/pkill", "x");
    unveil("/usr/sbin/openssl", "x");      unveil("/usr/local/bin/curl", "x");
    unveil("/usr/local/bin/tor", "x");     unveil("/usr/local/bin/gpg", "x");
    unveil("/usr/local/bin/espeak", "x");

    if(pledge("stdio rpath wpath cpath proc exec tty inet unveil", NULL) == -1)
        err(1, "pledge");
    #endif

    localfilez();

    #ifdef openbsd /* after localfilez() need path vars loaded */
    jail = (const char *)tcdir;
    unveil(jail, "rwxc");
    unveil(NULL, NULL);
    #endif

    createtcdirs();
    FILE *f = NULL;
    f = fopen(tcrc, "r"); (f == NULL) ? createrc() : (void)fclose(f);
    f = fopen(tcconf, "r"); (f == NULL) ? createtcconf() : (void)fclose(f);

    f = fopen(tclock, "r");
    if(f != NULL) {

        char p[10] = {"\0"};
        fgets(p, 10, f);
        int run = kill(atoi(p), 0);
        if(run == 0) errx(1, "already running (pid: %s)", p);
        else {

            fclose(f);
            unlink(tclock);
            f = fopen(tclock, "w");
            fprintf(f, "%d", getpid());
            fclose(f);
        }
    }
    else {

        f = fopen(tclock, "w");
        fprintf(f, "%d", getpid());
        fclose(f);
    }

    fflush(f);
    f = fopen(tcconf, "r");
    if(f == NULL) errx(1, "can't open tc.conf");
    if(f != NULL) if(checkfsize(f, tcconf, MAX_FSIZE, 1) == 0) loadconfig();

    fflush(f);
    if((f = popen("gpg --list-secret-keys", "r")) == NULL)
        errx(1, "empty PGP keyring");
    else pclose(f);

    gettorver(); getgnupgver(); loadkeysid(); socks4astr();
    int i;
    while((i = getopt(argc, argv, "un:s:")) != -1) {

        switch(i) {

        case 'n': strlcpy(nick, optarg, N); chat(); exit(0);
        case 's': strlcpy(nick, optarg, N); sym = 1; chat(); exit(0);
        case 'u': update();
        default:  usage();
        }
    }
    usage();
    return(0);
}
